# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Test the outgoing runner."""

import logging
import tempfile
import unittest
import mailman.handlers.validate_authenticity

from dns.resolver import Timeout
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.handlers.validate_authenticity import ValidateAuthenticity
from mailman.testing.helpers import (
    specialized_message_from_string as message_from_string,
)
from mailman.testing.layers import ConfigLayer
from unittest.mock import patch


def get_mock_dnsfunc(mapping):
    """Mock dnsfunc to pass to load_pk_from_dns."""

    def mock_dnsfunc(dom, **kw):
        return mapping.get(dom)

    return mock_dnsfunc


class TestValidateAuthenticity(unittest.TestCase):
    """Test Authentication-Results generation."""
    layer = ConfigLayer

    def setUp(self):
        # stub dns
        dkim0 = ["v=DKIM1; k=rsa; p=MIIBIjANBgkqhkiG9w0BAQEFAAO"
                 "CAQ8AMIIBCgKCAQEAg1i2lO83x/r58cbo/JSBwfZrrct6S/"
                 "yi4L6GsG3wNgFE9lO3orzBwnAEJJM33WrvJfOWia1fAx64V"
                 "s1QEpYtLFCzyeIhDDMaHv/G8NgKPgnWK4gI8/x2Q2SYCmiq"
                 "il66oHaSOC2phMDRI+c/Q35MlZbc2FqlgevpKzdCg+YE6mY"
                 "A0XN7/tdQplbx4meLVsVPIL9QCP4yu8oBsNqcwyxkQafJuc"
                 "VyoZI+VEO+dySw3QXNdmJhr7y1hD1tCNqoAG0iphKQVXPXm"
                 "GnGhaxaVU92Kq5UKL6/LiTZ1piqyJfJyZ/zCgH+mtY8MNk9"
                 "f7LHpwFljI7TbYmr7MmV3d6xj3sghwIDAQAB"]
        dkim1 = ["v=DKIM1; k=rsa; p=MIGfMA0GCSqGSIb3DQEBAQUAA4GNA"
                 "DCBiQKBgQDkHlOQoBTzWRiGs5V6NpP3idY6Wk08a5qhdR6w"
                 "y5bdOKb2jLQiY/J16JYi0Qvx/byYzCNb3W91y3FutACDfzw"
                 "Q/BC/e/8uBsCR+yz1Lxj+PL6lHvqMKrM3rG4hstT5QjvHO9"
                 "PzoxZyVYLzBfO2EeC3Ip3G+2kryOTIKT+l/K4w3QIDAQAB"]
        dmarc = ["v=DMARC1; p=reject; rua=mailto:dmarc.reports@"
                 "valimail.com,mailto:dmarc_agg@vali.email; ruf="
                 "mailto:dmarc.reports@valimail.com,mailto:dmarc_"
                 "c0cb7153_afrf@vali.email"]

        records = {b"google2048._domainkey.valimail.com.": ''.join(dkim0),
                   b"_dmarc.valimail.com.": ''.join(dmarc),
                   b"dummy._domainkey.example.org.": ''.join(dkim1)}
        mailman.handlers.validate_authenticity.dnsfunc = get_mock_dnsfunc(
            records)
        self.keyfile = tempfile.NamedTemporaryFile(delete=True)

    def tearDown(self):
        self.keyfile.close()

    def test_chain_validation_pass(self):
        config.push('dkim_and_cv', """
        [ARC]
        enabled: yes
        authserv_id: lists.example.org
        dkim: no
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'dkim_and_cv')

        lst = create_list('test@example.com')
        msgdata = {}

        msg = message_from_string("""MIME-Version: 1.0
Return-Path: <jqd@d1.example.org>
ARC-Seal: a=rsa-sha256;
    b=dOdFEyhrk/tw5wl3vMIogoxhaVsKJkrkEhnAcq2XqOLSQhPpGzhGBJzR7k1sWGokon3TmQ
    7TX9zQLO6ikRpwd/pUswiRW5DBupy58fefuclXJAhErsrebfvfiueGyhHXV7C1LyJTztywzn
    QGG4SCciU/FTlsJ0QANrnLRoadfps=; cv=none; d=example.org; i=1; s=dummy;
    t=12345
ARC-Message-Signature: a=rsa-sha256;
    b=QsRzR/UqwRfVLBc1TnoQomlVw5qi6jp08q8lHpBSl4RehWyHQtY3uOIAGdghDk/mO+/Xpm
    9JA5UVrPyDV0f+2q/YAHuwvP11iCkBQkocmFvgTSxN8H+DwFFPrVVUudQYZV7UDDycXoM6UE
    cdfzLLzVNPOAHEDIi/uzoV4sUqZ18=;
    bh=KWSe46TZKCcDbH4klJPo+tjk5LWJnVRlP5pvjXFZYLQ=; c=relaxed/relaxed;
    d=example.org;
    h=from:to:date:subject:mime-version:arc-authentication-results;
    i=1; s=dummy; t=12345
ARC-Authentication-Results: i=1; lists.example.org;
    spf=pass smtp.mfrom=jqd@d1.example;
    dkim=pass (1024-bit key) header.i=@d1.example;
    dmarc=pass
Received: from segv.d1.example (segv.d1.example [72.52.75.15])
    by lists.example.org (8.14.5/8.14.5) with ESMTP id t0EKaNU9010123
    for <arc@example.org>; Thu, 14 Jan 2015 15:01:30 -0800 (PST)
    (envelope-from jqd@d1.example)
Authentication-Results: lists.example.org;
    spf=pass smtp.mfrom=jqd@d1.example;
    dkim=pass (1024-bit key) header.i=@d1.example;
    dmarc=pass
Received: by 10.157.14.6 with HTTP; Tue, 3 Jan 2017 12:22:54 -0800 (PST)
Message-ID: <54B84785.1060301@d1.example.org>
Date: Thu, 14 Jan 2015 15:00:01 -0800
From: John Q Doe <jqd@d1.example.org>
To: arc@dmarc.org
Subject: Example 1

Hey gang,
This is a test message.
--J.
""")
        ValidateAuthenticity().process(lst, msg, msgdata)

        res = ["lists.example.org; spf=pass smtp.mfrom=jqd@d1.example"
               "; dkim=pass header.i=@d1.example; dmarc=pass; arc=pass"]
        self.assertEqual(msg["Authentication-Results"], ''.join(res))

    def test_chain_validation_fail(self):
        # Set the log-level for dkimpy to CRITICAL To avoid error output from
        # logger.
        logging.getLogger('dkimpy').setLevel(logging.CRITICAL)
        config.push('dkim_and_cv', """
        [ARC]
        enabled: yes
        authserv_id: lists.example.org
        dkim: yes
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'dkim_and_cv')

        lst = create_list('test@example.com')
        msgdata = {}

        msg = message_from_string("""MIME-Version: 1.0
Return-Path: <jqd@d1.example.org>
ARC-Seal: a=rsa-sha256;
    b=dOdFEyhrk/tw5wl3vMIogoxhaVsKJkrkEhnAcq2XqOLSQhPpGzhGBJzR7k1sWGokon3TmQ
    7TX9zQLO6ikRpwd/pUswiRW5DBupy58fefuclXJAhErsrebfvfiueGyhHXV7C1LyJTztywzn
    QGG4SCciU/FTlsJ0QANrnLRoadfps=; cv=none; d=example.org; i=1; s=dummy;
    t=12345
ARC-Message-Signature: a=rsa-sha256;
    b=QsRzR/UqwRfVLBc1TnoQomlVw5qi6jp08q8lHpBSl4RehWyHQtY3uOIAGdghDk/mO+/Xpm
    9JA5UVrPyDV0f+2q/YAHuwvP11iCkBQkocmFvgTSxN8H+DwFFPrVVUudQYZV7UDDycXoM6UE
    cdfzLLzVNPOAHEDIi/uzoV4sUqZ18=;
    bh=KWSe46TZKCcDbH4klJPo+tjk5LWJnVRlP5pvjXFZYLQ=; c=relaxed/relaxed;
    d=example.org; h=from:to:date:subject:mime-version:arc-authentication-
    results; i=1; s=dummy; t=12345
ARC-Authentication-Results: i=1; lists.example.org;
    spf=pass smtp.mfrom=jqd@d1.example;
    dkim=pass (1024-bit key) header.i=@d1.example;
    dmarc=pass
Received: from segv.d1.example (segv.d1.example [72.52.75.15])
    by lists.example.org (8.14.5/8.14.5) with ESMTP id t0EKaNU9010123
    for <arc@example.org>; Thu, 14 Jan 2015 15:01:30 -0800 (PST)
    (envelope-from jqd@d1.example)
Authentication-Results: lists.example.org;
    spf=pass smtp.mfrom=jqd@d1.example;
    dkim=pass (1024-bit key) header.i=@d1.example;
    dmarc=pass
Received: by 10.157.14.6 with HTTP; Tue, 3 Jan 2017 12:22:54 -0800 (PST)
Message-ID: <54B84785.1060301@d1.example.org>
Date: Thu, 14 Jan 2015 15:00:01 -0800
From: John Q Doe <jqd@d1.example.org>
To: arc@dmarc.org
Subject: Example 1

Hey gang(modified),
This is a test message.
--J.
""")

        ValidateAuthenticity().process(lst, msg, msgdata)
        res = ['lists.example.org; spf=pass smtp.mfrom=jqd@d1.example',
               '; dkim=pass header.i=@d1.example; dmarc=pass; arc=fail ',
               '(Most recent ARC-Message-Signature did not validate)']
        self.assertEqual(msg["Authentication-Results"], ''.join(res))

    def test_authentication_whitelist_hit(self):
        config.push('just_dkim', """
        [ARC]
        enabled: yes
        authserv_id: example.com
        dkim: yes
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'just_dkim')

        lst = create_list('test@example.com')
        msgdata = {}

        msg = message_from_string(
            """DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed;
        d=valimail.com; s=google2048;
        h=mime-version:from:date:message-id:subject:to;
        bh=3VWGQGY+cSNYd1MGM+X6hRXU0stl8JCaQtl4mbX/j2I=;
        b=gntRk4rCVYIGkpO09ROkbs3n4YSIcp/Pi7tUnSIgs8uS+uZ2a77dG+/qlSvnk+mWET
         IBrkt1YpDzev/0ITTDy/zgTHjPiQIFcg9Q+3hn3sTz8ExCyM8/YYgoPqSs3oUXn3jwXk
         N/wpMuF29LTVp1gpkYzaoCDNPGd1Wag6Vh2lw65S7ruECCAdBm5XeSnvTOzIC0E/jmEt
         3hvaPiKAohCAsC5JAN89EATPOjnYJL4Q6X6p2qUsusz/8tkHuYvReHmxQkjQ0/N3fPP0
         6VfkIrPOHympq6qDUizbjiBmgiMWKnarrptblJvyt66/aIHx+QamP6LUA+/RUFY1q7TG
         MSDg==
Authentication-Results: example.com; spf=pass smtp.mailfrom=gmail.com
MIME-Version: 1.0
From: Gene Shuman <gene@valimail.com>
Date: Wed, 25 Jan 2017 16:13:31 -0800
Message-ID:
  <CANtLugNVcUMfjVH22FN=+A6Y_Ss+QX_=GnJ3xGfDY1iuEbbuRA@mail.gmail.com>
Subject: Test
To: geneshuman@gmail.com
Content-Type: text/plain; charset=UTF-8

This is a test!
""")

        ValidateAuthenticity().process(lst, msg, msgdata)

        res = ['example.com; spf=pass smtp.mailfrom=gmail.com',
               '; dkim=pass header.d=valimail.com; arc=none ',
               '(Message is not ARC signed)']
        self.assertEqual(msg["Authentication-Results"], ''.join(res))

    def test_authentication_whitelist_miss(self):
        config.push('just_dkim', """
        [ARC]
        enabled: yes
        authserv_id: test.com
        trusted_authserv_ids: example.com
        dkim: yes
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'just_dkim')

        lst = create_list('test@example.com')
        msgdata = {}

        msg = message_from_string(
            """DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed;
        d=valimail.com; s=google2048;
        h=mime-version:from:date:message-id:subject:to;
        bh=3VWGQGY+cSNYd1MGM+X6hRXU0stl8JCaQtl4mbX/j2I=;
        b=gntRk4rCVYIGkpO09ROkbs3n4YSIcp/Pi7tUnSIgs8uS+uZ2a77dG+/qlSvnk+mWET
         IBrkt1YpDzev/0ITTDy/zgTHjPiQIFcg9Q+3hn3sTz8ExCyM8/YYgoPqSs3oUXn3jwXk
         N/wpMuF29LTVp1gpkYzaoCDNPGd1Wag6Vh2lw65S7ruECCAdBm5XeSnvTOzIC0E/jmEt
         3hvaPiKAohCAsC5JAN89EATPOjnYJL4Q6X6p2qUsusz/8tkHuYvReHmxQkjQ0/N3fPP0
         6VfkIrPOHympq6qDUizbjiBmgiMWKnarrptblJvyt66/aIHx+QamP6LUA+/RUFY1q7TG
         MSDg==
Authentication-Results: example_no.com; spf=pass smtp.mailfrom=gmail.com
MIME-Version: 1.0
From: Gene Shuman <gene@valimail.com>
Date: Wed, 25 Jan 2017 16:13:31 -0800
Message-ID:
  <CANtLugNVcUMfjVH22FN=+A6Y_Ss+QX_=GnJ3xGfDY1iuEbbuRA@mail.gmail.com>
Subject: Test
To: geneshuman@gmail.com
Content-Type: text/plain; charset=UTF-8

This is a test!
""")

        ValidateAuthenticity().process(lst, msg, msgdata)
        self.assertEqual(msg["Authentication-Results"],
                         'test.com; dkim=pass header.d=valimail.com; arc=none '
                         '(Message is not ARC signed)')

    def test_authentication_bad_outlook_header(self):
        config.push('just_dkim', """
        [ARC]
        enabled: yes
        authserv_id: test.com
        trusted_authserv_ids: example.com
        dkim: yes
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'just_dkim')

        lst = create_list('test@example.com')
        msgdata = {}

        msg = message_from_string(
            """DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed;
        d=valimail.com; s=google2048;
        h=mime-version:from:date:message-id:subject:to;
        bh=3VWGQGY+cSNYd1MGM+X6hRXU0stl8JCaQtl4mbX/j2I=;
        b=gntRk4rCVYIGkpO09ROkbs3n4YSIcp/Pi7tUnSIgs8uS+uZ2a77dG+/qlSvnk+mWET
         IBrkt1YpDzev/0ITTDy/zgTHjPiQIFcg9Q+3hn3sTz8ExCyM8/YYgoPqSs3oUXn3jwXk
         N/wpMuF29LTVp1gpkYzaoCDNPGd1Wag6Vh2lw65S7ruECCAdBm5XeSnvTOzIC0E/jmEt
         3hvaPiKAohCAsC5JAN89EATPOjnYJL4Q6X6p2qUsusz/8tkHuYvReHmxQkjQ0/N3fPP0
         6VfkIrPOHympq6qDUizbjiBmgiMWKnarrptblJvyt66/aIHx+QamP6LUA+/RUFY1q7TG
         MSDg==
Authentication-Results: groups.example.com; dkim=none (message not signed)
 header.d=none;groups.groups.example.com; dmarc=none action=none
 header.from=AList.onmicrosoft.com;
MIME-Version: 1.0
From: A List <alist@example.com>
Date: Wed, 25 Jan 2017 16:13:31 -0800
Message-ID:
  <CANtLugNVcUMfjVH22FN=+A6Y_Ss+QX_=GnJ3xGfDY1iuEbbuRA@mail.gmail.com>
Subject: Test
To: geneshuman@gmail.com
Content-Type: text/plain; charset=UTF-8

This is a test!
""")

        ValidateAuthenticity().process(lst, msg, msgdata)
        self.assertEqual(msg["Authentication-Results"],
                         'test.com; dkim=fail header.d=valimail.com; arc=none '
                         '(Message is not ARC signed)')


class TestTimeout(unittest.TestCase):
    """Test socket.error occurring in the delivery function."""

    layer = ConfigLayer

    def setUp(self):

        class timeout():
            def __init__(self):
                self.counter = 0

            def __call__(self, *args, **kwargs):
                self.counter += 1
                raise Timeout

        self.mock_timeout = timeout()
        self.patcher = patch(
            'mailman.handlers.validate_authenticity.authenticate_message',
            side_effect=self.mock_timeout)
        self.patcher.start()
        self.keyfile = tempfile.NamedTemporaryFile(delete=True)

    def tearDown(self):
        self.patcher.stop()
        self.keyfile.close()

    def test_timeout(self):
        config.push('just_dkim', """
        [ARC]
        enabled: yes
        dkim: yes
        dmarc: no
        privkey: {}
        """.format(self.keyfile.name))
        self.addCleanup(config.pop, 'just_dkim')

        mlist = create_list('test@example.com')
        msg = """DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed;
        d=valimail.com; s=google2048;
        h=mime-version:from:date:message-id:subject:to;
        bh=3VWGQGY+cSNYd1MGM+X6hRXU0stl8JCaQtl4mbX/j2I=;
        b=gntRk4rCVYIGkpO09ROkbs3n4YSIcp/Pi7tUnSIgs8uS+uZ2a77dG+/qlSvnk+mWET
         IBrkt1YpDzev/0ITTDy/zgTHjPiQIFcg9Q+3hn3sTz8ExCyM8/YYgoPqSs3oUXn3jwXk
         N/wpMuF29LTVp1gpkYzaoCDNPGd1Wag6Vh2lw65S7ruECCAdBm5XeSnvTOzIC0E/jmEt
         3hvaPiKAohCAsC5JAN89EATPOjnYJL4Q6X6p2qUsusz/8tkHuYvReHmxQkjQ0/N3fPP0
         6VfkIrPOHympq6qDUizbjiBmgiMWKnarrptblJvyt66/aIHx+QamP6LUA+/RUFY1q7TG
         MSDg==
MIME-Version: 1.0
From: Gene Shuman <gene@valimail.com>
Date: Wed, 25 Jan 2017 16:13:31 -0800
Message-ID: <CANtLugNVcUMfjVH22FN=+A6Y_Ss+QX_=GnJ3xGfDY1iuEbbuRA@gmail.com>
Subject: Test
To: geneshuman@gmail.com
Content-Type: text/plain; charset=UTF-8

This is a test!
"""
        msg = message_from_string(msg)

        with self.assertRaises(Timeout):
            ValidateAuthenticity().process(mlist, msg, {})

        # Make sure that timeout was called twice.
        self.assertEqual(self.mock_timeout.counter, 2)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Model for members."""

from datetime import datetime
from mailman.core.constants import system_preferences
from mailman.database.model import Model
from mailman.database.transaction import dbconnection
from mailman.database.types import Enum, SAUnicode, UUID
from mailman.interfaces.action import Action
from mailman.interfaces.address import IAddress
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.member import (
    DeliveryStatus,
    IMember,
    IMembershipManager,
    MemberRole,
    MembershipError,
    SubscriptionMode,
    UnsubscriptionEvent,
)
from mailman.interfaces.user import IUser, UnverifiedAddressError
from mailman.interfaces.usermanager import IUserManager
from mailman.utilities.datetime import now
from mailman.utilities.uid import UIDFactory
from public import public
from sqlalchemy import and_, Column, DateTime, ForeignKey, Integer
from sqlalchemy.orm import relationship
from zope.component import getUtility
from zope.event import notify
from zope.interface import implementer


uid_factory = UIDFactory(context='members')


@public
@implementer(IMember)
class Member(Model):
    """See `IMember`."""

    __tablename__ = 'member'

    id = Column(Integer, primary_key=True)
    _member_id = Column(UUID)
    role = Column(Enum(MemberRole), index=True)
    list_id = Column(SAUnicode, index=True)

    moderation_action = Column(Enum(Action))

    address_id = Column(Integer, ForeignKey('address.id'), index=True)
    _address = relationship('Address')
    preferences_id = Column(Integer, ForeignKey('preferences.id'), index=True)
    preferences = relationship('Preferences')
    user_id = Column(Integer, ForeignKey('user.id'), index=True)
    _user = relationship('User')

    bounce_score = Column(Integer, default=0, index=True)
    last_bounce_received = Column(DateTime)
    last_warning_sent = Column(DateTime, default=datetime.min, index=True)
    total_warnings_sent = Column(Integer, default=0, index=True)

    _mailing_list = None

    def __init__(self, role, list_id, subscriber):
        self._member_id = uid_factory.new()
        self.role = role
        self.list_id = list_id
        if IAddress.providedBy(subscriber):
            self._address = subscriber
            # Look this up dynamically.
            self._user = None
        elif IUser.providedBy(subscriber):
            self._user = subscriber
            # Look this up dynamically.
            self._address = None
        else:
            raise ValueError('subscriber must be a user or address')
        if role in (MemberRole.owner, MemberRole.moderator):
            self.moderation_action = Action.accept
        else:
            assert role in (MemberRole.member, MemberRole.nonmember), (
                'Invalid MemberRole: {}'.format(role))
            self.moderation_action = None

    def __repr__(self):
        return '<Member: {} on {} as {}>'.format(
            self.address, self.mailing_list.fqdn_listname, self.role)

    @property
    def mailing_list(self):
        """See `IMember`."""
        if self._mailing_list is None:
            self._mailing_list = getUtility(
                IListManager).get_by_list_id(self.list_id)
        return self._mailing_list

    @property
    def member_id(self):
        """See `IMember`."""
        return self._member_id

    @property
    def address(self):
        """See `IMember`."""
        return (self._user.preferred_address
                if self._address is None
                else self._address)

    @property
    def subscription_mode(self):
        """See `IMember`"""
        return (SubscriptionMode.as_address
                if self._address
                else SubscriptionMode.as_user)

    @address.setter
    def address(self, new_address):
        """See `IMember`."""
        if self._address is None:
            # XXX Either we need a better exception here, or we should allow
            # changing a subscription from preferred address to explicit
            # address (and vice versa via del'ing the .address attribute.
            raise MembershipError('Membership is via preferred address')
        if new_address.verified_on is None:
            # A member cannot change their subscription address to an
            # unverified address.
            raise UnverifiedAddressError('Unverified address')
        user = getUtility(IUserManager).get_user(new_address.email)
        if user is None or user != self.user:
            raise MembershipError('Address is not controlled by user')
        self._address = new_address

    @property
    def user(self):
        """See `IMember`."""
        return (self._user
                if self._address is None
                else self._address.user)

    @property
    def subscriber(self):
        return (self._user if self._address is None else self._address)

    @property
    def display_name(self):
        # Try to find a non-empty display name.  We first look at the directly
        # subscribed record, which will either be the address or the user.
        # That's handled automatically by going through member.subscriber.  If
        # that doesn't give us something useful, try whatever user is linked
        # to the subscriber.
        if self.subscriber.display_name:
            return self.subscriber.display_name
        # If an unlinked address is subscribed there will be no .user.
        elif self.user is not None and self.user.display_name:
            return self.user.display_name
        else:
            return ''

    def _lookup(self, preference, default=None):
        pref = getattr(self.preferences, preference)
        if pref is not None:
            return pref
        pref = getattr(self.address.preferences, preference)
        if pref is not None:
            return pref
        if self.address.user:
            pref = getattr(self.address.user.preferences, preference)
            if pref is not None:
                return pref
        if default is None:
            return getattr(system_preferences, preference)
        return default

    @property
    def acknowledge_posts(self):
        """See `IMember`."""
        return self._lookup('acknowledge_posts')

    @property
    def preferred_language(self):
        """See `IMember`."""
        missing = object()
        language = self._lookup('preferred_language', missing)
        return (self.mailing_list.preferred_language
                if language is missing
                else language)

    @property
    def receive_list_copy(self):
        """See `IMember`."""
        return self._lookup('receive_list_copy')

    @property
    def receive_own_postings(self):
        """See `IMember`."""
        return self._lookup('receive_own_postings')

    @property
    def delivery_mode(self):
        """See `IMember`."""
        return self._lookup('delivery_mode')

    @property
    def delivery_status(self):
        """See `IMember`."""
        return self._lookup('delivery_status')

    @dbconnection
    def unsubscribe(self, store):
        """See `IMember`."""
        # Yes, this must get triggered before self is deleted.
        notify(UnsubscriptionEvent(self.mailing_list, self))
        store.delete(self.preferences)
        store.delete(self)


@public
@implementer(IMembershipManager)
class MembershipManager:
    """See `IMembershipManager`."""

    @dbconnection
    def memberships_pending_warning(self, store):
        """See `IMembershipManager`."""
        from mailman.model.mailinglist import MailingList
        from mailman.model.preferences import Preferences

        # maxking: We don't care so much about the bounce score here since it
        # could have been reset due to bounce info getting stale. We will send
        # warnings to people who have been disabled already, regardless of
        # their bounce score. Same is true below for removal.
        query = store.query(
            Member,
            MailingList.bounce_you_are_disabled_warnings_interval).join(
            MailingList, Member.list_id == MailingList._list_id).join(
            Member.preferences).filter(and_(
                Member.role == MemberRole.member,
                MailingList.process_bounces == True,       # noqa: E712
                Member.total_warnings_sent < MailingList.bounce_you_are_disabled_warnings,  # noqa: E501
                Preferences.delivery_status == DeliveryStatus.by_bounces))

        # XXX(maxking): This is IMO a query that *should* work, but I haven't
        # been able to get it to work in my tests. It could be due to lack of
        # native datetime type in SQLite, which we use for tests. Hence, we are
        # going to do some filtering in Python, which is inefficient, but
        # works. We do filter as much as possible in SQL, so hopefully the
        # output of the above query won't be *too* many.
        # TODO: Port the Python loop below to SQLAlchemy.

        # (func.DATETIME(Member.last_warning_sent) +
        # func.DATETIME(MailingList.bounce_you_are_disabled_warnings_interval))
        # < func.DATETIME(now())))

        for member, interval in query.all():
            if (member.last_warning_sent + interval) <= now():
                yield member

    @dbconnection
    def memberships_pending_removal(self, store):
        """See `IMembershipManager`."""
        from mailman.model.mailinglist import MailingList
        from mailman.model.preferences import Preferences

        query = store.query(
            Member,
            MailingList.bounce_you_are_disabled_warnings_interval,
            MailingList.bounce_you_are_disabled_warnings).join(
            MailingList, Member.list_id == MailingList._list_id).join(
            Member.preferences).filter(and_(
                Member.role == MemberRole.member,
                MailingList.process_bounces == True,    # noqa: E712
                Member.total_warnings_sent >= MailingList.bounce_you_are_disabled_warnings,     # noqa: E501
                Preferences.delivery_status == DeliveryStatus.by_bounces))

        for member, interval, warnings in query.all():
            if (member.last_warning_sent + interval) <= now() or warnings == 0:
                yield member



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interfaces describing the state of a workflow."""

from public import public
from zope.interface import Attribute, Interface


@public
class IWorkflowState(Interface):
    """The state of a workflow."""

    token = Attribute('A unique key identifying the workflow instance.')

    step = Attribute("This workflow's next step.")

    data = Attribute('Additional data (may be JSON-encoded).')


@public
class IWorkflowStateManager(Interface):
    """The workflow states manager."""

    def save(token, step, data=None):
        """Save the state of a workflow.

        :param token: A unique token identifying this workflow instance.
        :type token: str
        :param step: The next step for this workflow.
        :type step: str
        :param data: Additional data (workflow-specific).
        :type data: str
        """

    def restore(token):
        """Get the saved state for a workflow or None if nothing was saved.

        :param token: A unique token identifying this workflow instance.
        :type token: str
        :return: The saved state associated with this name/token pair, or None
            if the pair isn't in the database.
        :rtype: ``IWorkflowState``
        """

    def discard(token):
        """Throw away the saved state for a workflow.

        :param token: A unique token identifying this workflow instance.
        :type token: str
        """

    def get_all_tokens():
        """Get a generator yielding all saved workflow tokens.

        :return: A generator yielding all saved workflow tokens.
        :rtype: string
        """

    count = Attribute('The number of saved workflows in the database.')



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Application support for chain processing."""

from mailman.config import config
from mailman.interfaces.chain import IChain, LinkAction
from mailman.utilities.modules import add_components
from public import public


@public
def process(mlist, msg, msgdata, start_chain='default-posting-chain'):
    """Process the message through a chain.

    :param mlist: the IMailingList for this message.
    :param msg: The Message object.
    :param msgdata: The message metadata dictionary.
    :param start_chain: The name of the chain to start the processing with.
    """
    # Set up some bookkeeping.
    chain_stack = []
    msgdata['rule_hits'] = hits = []
    msgdata['rule_misses'] = misses = []
    # Find the starting chain and begin iterating through its links.
    chain = config.chains[start_chain]
    chain_iter = chain.get_links(mlist, msg, msgdata)
    # Loop until we've reached the end of all processing chains.
    while chain:
        # Iterate over all links in the chain.  Do this outside a for-loop so
        # we can capture a chain's link iterator in mid-flight.  This supports
        # the 'detour' link action.
        try:
            link = next(chain_iter)
        except StopIteration:
            # This chain is exhausted.  Pop the last chain on the stack and
            # continue iterating through it.  If there's nothing left on the
            # chain stack then we're completely finished processing.
            if len(chain_stack) == 0:
                return
            chain, chain_iter = chain_stack.pop()
            continue
        if link.rule.check(mlist, msg, msgdata):
            if link.rule.record:
                hits.append(link.rule.name)
            # The rule matched so run its action.
            if link.action is LinkAction.jump:
                chain = link.chain
                chain_iter = chain.get_links(mlist, msg, msgdata)
                continue
            elif link.action is LinkAction.detour:
                # Push the current chain so that we can return to it when
                # the next chain is finished.
                chain_stack.append((chain, chain_iter))
                chain = link.chain
                chain_iter = chain.get_links(mlist, msg, msgdata)
                continue
            elif link.action is LinkAction.stop:
                # Stop all processing.
                return
            elif link.action is LinkAction.defer:
                # Just process the next link in the chain.
                pass
            elif link.action is LinkAction.run:
                link.function(mlist, msg, msgdata)
            else:
                raise AssertionError(
                    'Bad link action: {}'.format(link.action))
        else:
            # The rule did not match; keep going.
            if link.rule.record:
                misses.append(link.rule.name)


@public
def initialize():
    """Set up chains, both built-in and from the database."""
    add_components('chains', IChain, config.chains)



"""dmarc_attributes

Revision ID: 3002bac0c25a
Revises: a46993b05703
Create Date: 2016-10-30 22:05:17.881880

"""

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import exists_in_db
from mailman.database.types import Enum, SAUnicodeLarge
from mailman.interfaces.mailinglist import DMARCMitigateAction


# revision identifiers, used by Alembic.
revision = '3002bac0c25a'
down_revision = 'a46993b05703'


def upgrade():
    if not exists_in_db(op.get_bind(),
                        'mailinglist',
                        'dmarc_mitigate_action'
                        ):
        # SQLite may not have removed it when downgrading.  It should be OK
        # to just test one.
        op.add_column('mailinglist', sa.Column(
            'dmarc_mitigate_action',
            Enum(DMARCMitigateAction),
            nullable=True))
        op.add_column('mailinglist', sa.Column(
            'dmarc_mitigate_unconditionally',
            sa.Boolean,
            nullable=True))
        op.add_column('mailinglist', sa.Column(
            'dmarc_moderation_notice',
            SAUnicodeLarge(),
            nullable=True))
        op.add_column('mailinglist', sa.Column(
            'dmarc_wrapped_message_text',
            SAUnicodeLarge(),
            nullable=True))
    # Now migrate the data.  Don't import the table definition from the
    # models, it may break this migration when the model is updated in the
    # future (see the Alembic doc).
    mlist = sa.sql.table(
        'mailinglist',
        sa.sql.column('dmarc_mitigate_action', Enum(DMARCMitigateAction)),
        sa.sql.column('dmarc_mitigate_unconditionally', sa.Boolean),
        sa.sql.column('dmarc_moderation_notice', SAUnicodeLarge()),
        sa.sql.column('dmarc_wrapped_message_text', SAUnicodeLarge()),
        )
    # These are all new attributes so just set defaults.
    op.execute(mlist.update().values(dict(
        dmarc_mitigate_action=op.inline_literal(
            DMARCMitigateAction.no_mitigation),
        dmarc_mitigate_unconditionally=op.inline_literal(False),
        dmarc_moderation_notice=op.inline_literal(''),
        dmarc_wrapped_message_text=op.inline_literal(''),
        )))


def downgrade():
    with op.batch_alter_table('mailinglist') as batch_op:
        batch_op.drop_column('dmarc_mitigate_action')
        batch_op.drop_column('dmarc_mitigate_unconditionally')
        batch_op.drop_column('dmarc_moderation_notice')
        batch_op.drop_column('dmarc_wrapped_message_text')



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The process runner base class."""

import time
import signal
import logging
import traceback

from contextlib import suppress
from io import StringIO
from lazr.config import as_boolean, as_timedelta
from mailman.config import config
from mailman.core.i18n import _
from mailman.core.logging import reopen
from mailman.core.switchboard import Switchboard
from mailman.interfaces.languages import ILanguageManager
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.runner import (
    IRunner,
    RunnerCrashEvent,
    RunnerInterrupt,
)
from mailman.utilities.string import expand
from public import public
from zope.component import getUtility
from zope.event import notify
from zope.interface import implementer


dlog = logging.getLogger('mailman.debug')
elog = logging.getLogger('mailman.error')
rlog = logging.getLogger('mailman.runner')


@public
@implementer(IRunner)
class Runner:
    is_queue_runner = True

    def __init__(self, name, slice=None):
        """Create a runner.

        :param slice: The slice number for this runner.  This is passed
            directly to the underlying `ISwitchboard` object.  This is ignored
            for runners that don't manage a queue.
        :type slice: int or None
        """
        # Grab the configuration section.
        self.name = name
        section = getattr(config, 'runner.' + name)
        substitutions = config.paths
        substitutions['name'] = name
        numslices = int(section.instances)
        # Check whether the runner is queue runner or not; non-queue runner
        # should not have queue_directory or switchboard instance.
        if self.is_queue_runner:
            self.queue_directory = expand(section.path, None, substitutions)
            self.switchboard = Switchboard(
                name, self.queue_directory, slice, numslices, True)
        else:
            self.queue_directory = None
            self.switchboard = None
        self.sleep_time = as_timedelta(section.sleep_time)
        # sleep_time is a timedelta; turn it into a float for time.sleep().
        self.sleep_float = (86400 * self.sleep_time.days +
                            self.sleep_time.seconds +
                            self.sleep_time.microseconds / 1.0e6)
        self.max_restarts = int(section.max_restarts)
        self.start = as_boolean(section.start)
        self._stop = False
        self.status = 0

    def __repr__(self):
        return '<{} at {:#x}>'.format(self.__class__.__name__, id(self))

    def signal_handler(self, signum, frame):        # pragma: nocover
        signame = {
            signal.SIGTERM: 'SIGTERM',
            signal.SIGINT: 'SIGINT',
            signal.SIGUSR1: 'SIGUSR1',
            }.get(signum, signum)
        if signum == signal.SIGHUP:
            reopen()
            rlog.info('{} runner caught SIGHUP.  Reopening logs.'.format(
                self.name))
        elif signum in (signal.SIGTERM, signal.SIGINT, signal.SIGUSR1):
            self.stop()
            self.status = signum
            rlog.info('{} runner caught {}.  Stopping.'.format(
                self.name, signame))
            # As of Python 3.5, PEP 475 gets in our way.  Runners with long
            # time.sleep()'s in their _snooze() method (e.g. the retry runner)
            # will have their system call implemented time.sleep()
            # automatically retried at the C layer.  The only reliable way to
            # prevent this is to raise an exception in the signal handler.  The
            # standard run() method automatically suppresses this exception,
            # meaning, it's caught and ignored, but effectively breaks the
            # run() loop, which is just what we want.  Runners which implement
            # their own run() method must be prepared to catch
            # RunnerInterrupts, usually also ignoring them.
            raise RunnerInterrupt

    def set_signals(self):
        """See `IRunner`."""
        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGUSR1, self.signal_handler)

    def stop(self):
        """See `IRunner`."""
        self._stop = True

    def run(self):
        """See `IRunner`."""
        # Start the main loop for this runner.
        with suppress(KeyboardInterrupt, RunnerInterrupt):
            while True:
                # Once through the loop that processes all the files in the
                # queue directory.
                filecnt = self._one_iteration()
                # Do the periodic work for the subclass.
                self._do_periodic()
                # If the stop flag is set, we're done.
                if self._stop:
                    break
                # Give the runner an opportunity to snooze for a while, but
                # pass it the file count so it can decide whether to do more
                # work now or not.
                self._snooze(filecnt)
        self._clean_up()

    def _one_iteration(self):
        """See `IRunner`."""
        me = self.__class__.__name__
        dlog.debug('[%s] starting oneloop', me)
        # List all the files in our queue directory.  The switchboard is
        # guaranteed to hand us the files in FIFO order.
        if self.switchboard is None:
            files = []
        else:
            files = self.switchboard.files
        for filebase in files:
            dlog.debug('[%s] processing filebase: %s', me, filebase)
            try:
                # Ask the switchboard for the message and metadata objects
                # associated with this queue file.
                msg, msgdata = self.switchboard.dequeue(filebase)
            except Exception as error:
                # This used to just catch email.Errors.MessageParseError, but
                # other problems can occur in message parsing, e.g.
                # ValueError, and exceptions can occur in unpickling too.  We
                # don't want the runner to die, so we just log and skip this
                # entry, but preserve it for analysis.
                self._log(error)
                elog.error('Skipping and preserving unparseable message: %s',
                           filebase)
                self.switchboard.finish(filebase, preserve=True)
                config.db.abort()
                continue
            try:
                dlog.debug('[%s] processing onefile', me)
                self._process_one_file(msg, msgdata)
                dlog.debug('[%s] finishing filebase: %s', me, filebase)
                self.switchboard.finish(filebase)
            except Exception as error:
                # All runners that implement _dispose() must guarantee that
                # exceptions are caught and dealt with properly.  Still, there
                # may be a bug in the infrastructure, and we do not want those
                # to cause messages to be lost.  Any uncaught exceptions will
                # cause the message to be stored in the shunt queue for human
                # intervention.
                self._log(error)
                # Put a marker in the metadata for unshunting.
                msgdata['whichq'] = self.switchboard.name
                # It is possible that shunting can throw an exception, e.g. a
                # permissions problem or a MemoryError due to a really large
                # message.  Try to be graceful.
                try:
                    shunt = config.switchboards['shunt']
                    new_filebase = shunt.enqueue(msg, msgdata)
                    elog.error('SHUNTING: %s', new_filebase)
                    self.switchboard.finish(filebase)
                except Exception as error:
                    # The message wasn't successfully shunted.  Log the
                    # exception and try to preserve the original queue entry
                    # for possible analysis.
                    self._log(error)
                    elog.error(
                        'SHUNTING FAILED, preserving original entry: %s',
                        filebase)
                    self.switchboard.finish(filebase, preserve=True)
                config.db.abort()
            # Other work we want to do each time through the loop.
            dlog.debug('[%s] doing periodic', me)
            self._do_periodic()
            dlog.debug('[%s] committing transaction', me)
            config.db.commit()
            dlog.debug('[%s] checking short circuit', me)
            if self._short_circuit():
                dlog.debug('[%s] short circuiting', me)
                break
        dlog.debug('[%s] ending oneloop: %s', me, len(files))
        return len(files)

    def _process_one_file(self, msg, msgdata):
        """See `IRunner`."""
        # Do some common sanity checking on the message metadata.  It's got to
        # be destined for a particular mailing list.  This switchboard is used
        # to shunt off badly formatted messages.  We don't want to just trash
        # them because they may be fixable with human intervention.  Just get
        # them out of our sight.
        #
        # Find out which mailing list this message is destined for.
        mlist = None
        missing = object()
        # First try to dig out the target list by id.  If there's no list-id
        # in the metadata, fall back to the fqdn list name for backward
        # compatibility.
        list_manager = getUtility(IListManager)
        list_id = msgdata.get('listid', missing)
        fqdn_listname = None
        if list_id is missing:
            fqdn_listname = msgdata.get('listname', missing)
            # XXX Deprecate.
            if fqdn_listname is not missing:
                mlist = list_manager.get(fqdn_listname)
        else:
            mlist = list_manager.get_by_list_id(list_id)
        if mlist is None:
            identifier = (list_id if list_id is not None else fqdn_listname)
            elog.error(
                '%s runner "%s" shunting message for missing list: %s',
                msg['message-id'], self.name, identifier)
            config.switchboards['shunt'].enqueue(msg, msgdata)
            return
        # Now process this message.  We also want to set up the language
        # context for this message.  The context will be the preferred
        # language for the user if the sender is a member of the list, or it
        # will be the list's preferred language.  However, we must take
        # special care to reset the defaults, otherwise subsequent messages
        # may be translated incorrectly.
        if mlist is None:
            language_manager = getUtility(ILanguageManager)
            language = language_manager[config.mailman.default_language]
        elif msg.sender:
            member = mlist.members.get_member(msg.sender)
            language = (member.preferred_language
                        if member is not None
                        else mlist.preferred_language)
        else:
            language = mlist.preferred_language
        with _.using(language.code):
            msgdata['lang'] = language.code
            try:
                keepqueued = self._dispose(mlist, msg, msgdata)
            except Exception as error:
                # Trigger the Zope event and re-raise
                notify(RunnerCrashEvent(self, mlist, msg, msgdata, error))
                raise
        if keepqueued:
            self.switchboard.enqueue(msg, msgdata)

    def _log(self, exc):
        elog.error('Uncaught runner exception: %s', exc)
        s = StringIO()
        traceback.print_exc(file=s)
        elog.error('%s', s.getvalue())

    def _clean_up(self):
        """See `IRunner`."""
        pass

    def _dispose(self, mlist, msg, msgdata):
        """See `IRunner`."""
        raise NotImplementedError

    def _do_periodic(self):
        """See `IRunner`."""
        pass

    def _snooze(self, filecnt):
        """See `IRunner`."""
        if filecnt or self.sleep_float <= 0:
            return
        time.sleep(self.sleep_float)

    def _short_circuit(self):
        """See `IRunner`."""
        return self._stop



# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `mailman members` command."""

import unittest

from click.testing import CliRunner
from mailman.app.lifecycle import create_list
from mailman.commands.cli_findmember import findmember
from mailman.interfaces.member import MemberRole
from mailman.testing.helpers import subscribe
from mailman.testing.layers import ConfigLayer


class TestCLIFindMember(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._mlistb = create_list('bee@example.com')
        self._command = CliRunner()

    def test_basic_find(self):
        # Test a simple find of one membership.
        subscribe(self._mlist, 'Anne')
        result = self._command.invoke(findmember, ('.',))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Email: aperson@example.com
    List: ant.example.com
        MemberRole.member
""")

    def test_no_role(self):
        # Test for no matching roles.
        subscribe(self._mlist, 'Anne')
        result = self._command.invoke(findmember, ('--role', 'owner', '.',))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, '')

    def test_no_pattern_match(self):
        # test for no address matching pattern.
        subscribe(self._mlist, 'Anne')
        result = self._command.invoke(findmember, ('doesnt_match',))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, '')

    def test_case_insensitive_pattern(self):
        # Test that patterns are case insensitive.
        subscribe(self._mlist, 'Anne')
        result = self._command.invoke(findmember, ('APerson',))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Email: aperson@example.com
    List: ant.example.com
        MemberRole.member
""")

    def test_user_and_address(self):
        # Test finding members subscribed as user and as address.
        subscribe(self._mlist, 'Anne', as_user=True)
        subscribe(self._mlistb, 'Bart', as_user=False)
        result = self._command.invoke(findmember, ('.'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Email: aperson@example.com
    List: ant.example.com
        MemberRole.member
Email: bperson@example.com
    List: bee.example.com
        MemberRole.member
""")

    def test_only_role(self):
        # Test only finding requested role.
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlistb, 'Bart', role=MemberRole.owner)
        result = self._command.invoke(findmember, ('--role', 'owner', '.'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Email: bperson@example.com
    List: bee.example.com
        MemberRole.owner
""")

    def test_find_only_admins(self):
        # Test the administrators role.
        subscribe(self._mlist, 'Anne')
        subscribe(self._mlistb, 'Bart', role=MemberRole.owner)
        subscribe(self._mlist, 'Cate', role=MemberRole.moderator)
        subscribe(self._mlistb, 'Doug', role=MemberRole.nonmember)
        result = self._command.invoke(findmember, ('--role',
                                                   'administrator',
                                                   '.'))
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, """\
Email: bperson@example.com
    List: bee.example.com
        MemberRole.owner
Email: cperson@example.com
    List: ant.example.com
        MemberRole.moderator
""")



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Digest runner."""

import os
import re
import logging

from copy import deepcopy
from email.header import Header
from email.mime.message import MIMEMessage
from email.mime.text import MIMEText
from email.utils import formatdate, getaddresses, make_msgid
from io import StringIO
from mailman.config import config
from mailman.core.i18n import _
from mailman.core.runner import Runner
from mailman.email.message import Message, MultipartDigestMessage
from mailman.handlers.decorate import decorate
from mailman.interfaces.member import DeliveryMode, DeliveryStatus
from mailman.interfaces.template import ITemplateLoader
from mailman.utilities.mailbox import Mailbox
from mailman.utilities.scrubber import scrub
from mailman.utilities.string import expand, oneline, wrap
from public import public
from zope.component import getUtility


log = logging.getLogger('mailman.error')


class Digester:
    """Base digester class."""

    def __init__(self, mlist, volume, digest_number):
        self._mlist = mlist
        self._charset = mlist.preferred_language.charset
        # This will be used in the Subject, so use $-strings.
        self._digest_id = _(
            '${mlist.display_name} Digest, Vol ${volume}, Issue '
            '${digest_number}')
        self._subject = Header(self._digest_id,
                               self._charset,
                               header_name='Subject')
        self._message = self._make_message()
        self._digest_part = self._make_digest_part()
        self._message['From'] = mlist.request_address
        self._message['Subject'] = self._subject
        self._message['To'] = mlist.posting_address
        self._message['Reply-To'] = mlist.posting_address
        self._message['Date'] = formatdate(localtime=True)
        self._message['Message-ID'] = make_msgid()
        # In the rfc1153 digest, the masthead contains the digest boilerplate
        # plus any digest header.  In the MIME digests, the masthead and
        # digest header are separate MIME subobjects.  In either case, it's
        # the first thing in the digest, and we can calculate it now, so go
        # ahead and add it now.
        template = getUtility(ITemplateLoader).get(
            'list:member:digest:masthead', mlist)
        self._masthead = wrap(expand(template, mlist, dict(
            # For backward compatibility.
            got_list_email=mlist.posting_address,
            got_request_email=mlist.request_address,
            got_owner_email=mlist.owner_address,
            )))
        # Set things up for the table of contents.
        self._header = decorate('list:member:digest:header', mlist)
        self._toc = StringIO()
        print(_("Today's Topics:\n"), file=self._toc)

    def add_to_toc(self, msg, count):
        """Add a message to the table of contents."""
        subject = msg.get('subject', _('(no subject)'))
        subject = oneline(subject, in_unicode=True)
        # Don't include the redundant subject prefix in the toc
        mo = re.match('(re:? *)?({0})'.format(
            re.escape(self._mlist.subject_prefix)),
                      subject, re.IGNORECASE)
        if mo:
            subject = subject[:mo.start(2)] + subject[mo.end(2):]
        # Take only the first author we find.
        username = ''
        addresses = getaddresses(
            [oneline(msg.get('from', ''), in_unicode=True)])
        if addresses:
            username = addresses[0][0]
            if not username:
                username = addresses[0][1]
        if username:
            username = ' ({})'.format(username)
        lines = wrap('{:2}. {}'. format(count, subject), 65).split('\n')
        # See if the user's name can fit on the last line
        if len(lines[-1]) + len(username) > 70:
            lines.append(username)
        else:
            lines[-1] += username
        # Add this subject to the accumulating topics
        first = True
        for line in lines:
            if first:
                print(' ', line, file=self._toc)
                first = False
            else:
                print('     ', line.lstrip(), file=self._toc)

    def add_message(self, msg, count):
        """Add the message to the digest."""
        # We do not want all the headers of the original message to leak
        # through in the digest messages.
        keepers = {}
        for header in self._keepers:
            keepers[header] = msg.get_all(header, [])
        # Remove all the unkempt <wink> headers.  Use .keys() to allow for
        # destructive iteration...
        for header in msg.keys():
            del msg[header]
        # ... and add them in the designated order.
        for header in self._keepers:
            for value in keepers[header]:
                msg[header] = value
        # Add some useful extra stuff.
        msg['Message'] = count.decode('utf-8')


class MIMEDigester(Digester):
    """A MIME digester."""

    def __init__(self, mlist, volume, digest_number):
        super().__init__(mlist, volume, digest_number)
        masthead = MIMEText(self._masthead.encode(self._charset),
                            _charset=self._charset)
        masthead['Content-Description'] = self._subject
        self._message.attach(masthead)
        # Add the optional digest header.
        if len(self._header) > 0:
            header = MIMEText(self._header.encode(self._charset),
                              _charset=self._charset)
            header['Content-Description'] = _('Digest Header')
            self._message.attach(header)
        # Calculate the set of headers we're to keep in the MIME digest.
        self._keepers = set(config.digests.mime_digest_keep_headers.split())

    def _make_message(self):
        return MultipartDigestMessage('mixed')

    def _make_digest_part(self):
        return MultipartDigestMessage('digest')

    def add_toc(self, count):
        """Add the table of contents."""
        toc_text = self._toc.getvalue()
        try:
            toc_part = MIMEText(toc_text.encode(self._charset),
                                _charset=self._charset)
        except UnicodeError:
            toc_part = MIMEText(toc_text.encode('utf-8'), _charset='utf-8')
        toc_part['Content-Description'] = _(
            "Today's Topics (${count} messages)")
        self._message.attach(toc_part)

    def add_message(self, msg, count):
        """Add the message to the digest."""
        # Make a copy of the message object, since the RFC 1153 processing
        # scrubs out attachments.
        digest_msg = MIMEMessage(deepcopy(msg))
        digest_msg_content = digest_msg.get_payload(0)
        # It would be nice to add Message: n near the beginning, but there's no
        # method for that.  MUAs mostly don't display it anyway, so it doesn't
        # really matter.
        digest_msg_content['Message'] = str(count)
        self._digest_part.attach(digest_msg)

    def finish(self):
        """Finish up the digest, producing the email-ready copy."""
        self._message.attach(self._digest_part)
        footer_text = decorate('list:member:digest:footer', self._mlist)
        if len(footer_text) > 0:
            footer = MIMEText(footer_text.encode(self._charset),
                              _charset=self._charset)
            footer['Content-Description'] = _('Digest Footer')
            self._message.attach(footer)
        # This stuff is outside the normal MIME goo, and it's what the old
        # MIME digester did.  No one seemed to complain, probably because you
        # won't see it in an MUA that can't display the raw message.  We've
        # never got complaints before, but if we do, just wax this.  It's
        # primarily included for (marginally useful) backwards compatibility.
        self._message.postamble = _('End of ') + self._digest_id
        return self._message


class RFC1153Digester(Digester):
    """A digester of the format specified by RFC 1153."""

    def __init__(self, mlist, volume, digest_number):
        super().__init__(mlist, volume, digest_number)
        self._separator70 = '-' * 70
        self._separator30 = '-' * 30
        self._text = StringIO()
        print(self._masthead, file=self._text)
        print(file=self._text)
        # Add the optional digest header.
        if len(self._header) > 0:
            print(self._header, file=self._text)
            print(file=self._text)
        # Calculate the set of headers we're to keep in the RFC1153 digest.
        self._keepers = set(config.digests.plain_digest_keep_headers.split())

    def _make_message(self):
        return Message()

    def _make_digest_part(self):
        # Not actually used here but referenced in super().__init__().
        return self._message

    def add_toc(self, count):
        """Add the table of contents."""
        print(self._toc.getvalue(), file=self._text)
        print(file=self._text)
        print(self._separator70, file=self._text)
        print(file=self._text)

    def add_message(self, msg, count):
        """Add the message to the digest."""
        if count > 1:
            print(self._separator30, file=self._text)
            print(file=self._text)
        # Each message section contains a few headers.
        # add the Message: n header first.
        print('Message: {}'.format(count), file=self._text)
        # Then the others.
        for header in config.digests.plain_digest_keep_headers.split():
            if header in msg:
                value = oneline(msg[header], in_unicode=True)
                value = wrap('{}: {}'.format(header, value))
                value = '\n\t'.join(value.split('\n'))
                print(value, file=self._text)
        print(file=self._text)
        # Get the scrubbed payload.  This is the original payload with all
        # non text/plain parts replaced by notes that they've been removed.
        payload = scrub(msg)
        # Add the payload.
        print(payload, file=self._text)
        if not payload.endswith('\n'):
            print(file=self._text)

    def finish(self):
        """Finish up the digest, producing the email-ready copy."""
        footer_text = decorate('list:member:digest:footer', self._mlist)
        if len(footer_text) > 0:
            # MAS: There is no real place for the digest_footer in an RFC 1153
            # compliant digest, so add it as an additional message with
            # Subject: Digest Footer
            print(self._separator30, file=self._text)
            print(file=self._text)
            print('Subject: ' + _('Digest Footer'), file=self._text)
            print(file=self._text)
            print(footer_text, file=self._text)
            print(file=self._text)
            print(self._separator30, file=self._text)
            print(file=self._text)
        # Add the sign-off.
        sign_off = _('End of ') + self._digest_id
        print(sign_off, file=self._text)
        print('*' * len(sign_off), file=self._text)
        # If the digest message can't be encoded by the list character set,
        # fall back to utf-8 with error replacement.
        text = self._text.getvalue()
        try:
            self._message.set_payload(text.encode(self._charset),
                                      charset=self._charset)
        except UnicodeError:
            self._message.set_payload(text.encode('utf-8', errors='replace'),
                                      charset='utf-8')
        return self._message


@public
class DigestRunner(Runner):
    """The digest runner."""

    def _dispose(self, mlist, msg, msgdata):
        """See `IRunner`."""
        volume = msgdata['volume']
        digest_number = msgdata['digest_number']
        # Backslashes make me cry.
        code = mlist.preferred_language.code
        with Mailbox(msgdata['digest_path']) as mailbox, _.using(code):
            # Create the digesters.
            mime_digest = MIMEDigester(mlist, volume, digest_number)
            rfc1153_digest = RFC1153Digester(mlist, volume, digest_number)
            # Cruise through all the messages in the mailbox, first building
            # the table of contents and accumulating Subject: headers and
            # authors.  The question really is whether it's better from a1
            # performance and memory footprint to go through the mailbox once
            # and cache the messages in a list, or to cruise through the
            # mailbox twice.  We'll do the latter, but it's a complete guess.
            count = None
            for count, (key, message) in enumerate(mailbox.iteritems(), 1):
                mime_digest.add_to_toc(message, count)
                rfc1153_digest.add_to_toc(message, count)
            assert count is not None, 'No digest messages?'
            # Add the table of contents.
            mime_digest.add_toc(count)
            rfc1153_digest.add_toc(count)
            # Cruise through the set of messages a second time, adding them to
            # the actual digest.
            for count, (key, message) in enumerate(mailbox.iteritems(), 1):
                mime_digest.add_message(message, count)
                rfc1153_digest.add_message(message, count)
            # Finish up the digests.
            mime = mime_digest.finish()
            rfc1153 = rfc1153_digest.finish()
        # Calculate the recipients lists
        mime_recipients = set()
        rfc1153_recipients = set()
        # When someone turns off digest delivery, they will get one last
        # digest to ensure that there will be no gaps in the messages they
        # receive.
        digest_members = set(mlist.digest_members.members)
        for member in digest_members:
            if member.delivery_status is not DeliveryStatus.enabled:
                continue
            # Send the digest to the case-preserved address of the digest
            # members.
            email_address = member.address.original_email
            if member.delivery_mode == DeliveryMode.plaintext_digests:
                rfc1153_recipients.add(email_address)
            # We currently treat summary_digests the same as mime_digests.
            elif member.delivery_mode in (DeliveryMode.mime_digests,
                                          DeliveryMode.summary_digests):
                mime_recipients.add(email_address)
            else:
                raise AssertionError(
                    'Digest member "{}" unexpected delivery mode: {}'.format(
                        email_address, member.delivery_mode))
        # Add also the folks who are receiving one last digest.
        for address, delivery_mode in mlist.last_digest_recipients:
            if delivery_mode == DeliveryMode.plaintext_digests:
                rfc1153_recipients.add(address.original_email)
            # We currently treat summary_digests the same as mime_digests.
            elif delivery_mode in (DeliveryMode.mime_digests,
                                   DeliveryMode.summary_digests):
                mime_recipients.add(address.original_email)
            else:
                raise AssertionError(
                    'OLD recipient "{}" unexpected delivery mode: {}'.format(
                        address, delivery_mode))
        # Send the digests to the virgin queue for final delivery.
        queue = config.switchboards['virgin']
        if len(mime_recipients) > 0:
            queue.enqueue(mime,
                          recipients=mime_recipients,
                          listid=mlist.list_id,
                          isdigest=True)
        if len(rfc1153_recipients) > 0:
            queue.enqueue(rfc1153,
                          recipients=rfc1153_recipients,
                          listid=mlist.list_id,
                          isdigest=True)
        # Remove the digest mbox. (GL #259)
        os.remove(msgdata['digest_path'])



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The 'lists' subcommand."""

import sys
import click

from mailman.app.lifecycle import create_list, remove_list
from mailman.core.constants import system_preferences
from mailman.core.i18n import _
from mailman.database.transaction import transaction
from mailman.email.message import UserNotification
from mailman.interfaces.address import (
    IEmailValidator,
    InvalidEmailAddressError,
)
from mailman.interfaces.command import ICLISubCommand
from mailman.interfaces.domain import (
    BadDomainSpecificationError,
    IDomainManager,
)
from mailman.interfaces.languages import ILanguageManager
from mailman.interfaces.listmanager import IListManager, ListAlreadyExistsError
from mailman.interfaces.template import ITemplateLoader
from mailman.utilities.options import I18nCommand
from mailman.utilities.string import expand, wrap
from operator import attrgetter
from public import public
from zope.component import getUtility
from zope.interface import implementer


COMMASPACE = ', '


@click.command(
    cls=I18nCommand,
    help=_('List all mailing lists.'))
@click.option(
    '--advertised', '-a',
    is_flag=True, default=False,
    help=_('List only those mailing lists that are publicly advertised'))
@click.option(
    '--names/--no-names', '-n/-N',
    is_flag=True, default=False,
    help=_('Show also the list names'))
@click.option(
    '--descriptions/--no-descriptions', '-d/-D',
    is_flag=True, default=False,
    help=_('Show also the list descriptions'))
@click.option(
    '--quiet', '-q',
    is_flag=True, default=False,
    help=_('Less verbosity'))
@click.option(
    '--domain', 'domains',
    multiple=True, metavar='DOMAIN',
    help=_("""\
    List only those mailing lists hosted on the given domain, which
    must be the email host name.  Multiple -d options may be given.
    """))
@click.pass_context
def lists(ctx, advertised, names, descriptions, quiet, domains):
    mailing_lists = set()
    list_manager = getUtility(IListManager)
    # Gather the matching mailing lists.
    for mlist in list_manager.mailing_lists:
        if advertised and not mlist.advertised:
            continue
        if len(domains) > 0 and mlist.mail_host not in domains:
            continue
        mailing_lists.add(mlist)
    # Maybe no mailing lists matched.
    if len(mailing_lists) == 0:
        if not quiet:
            print(_('No matching mailing lists found'))
        sys.exit()
    count = len(mailing_lists)                  # noqa: F841
    if not quiet:
        print(_('${count} matching mailing lists found:'))
    # Calculate the longest identifier.
    longest = 0
    output = []
    for mlist in sorted(mailing_lists, key=attrgetter('list_id')):
        if names:
            identifier = '{} [{}]'.format(
                mlist.fqdn_listname, mlist.display_name)
        else:
            identifier = mlist.fqdn_listname
        longest = max(len(identifier), longest)
        output.append((identifier, mlist.description))
    # Print it out.
    if descriptions:
        format_string = '{0:{2}} - {1:{3}}'
    else:
        format_string = '{0:{2}}'
    for identifier, description in output:
        print(format_string.format(
            identifier, description, longest, 70 - longest))


@public
@implementer(ICLISubCommand)
class Lists:
    name = 'lists'
    command = lists


@click.command(
    cls=I18nCommand,
    help=_("""\
    Create a mailing list.

    The 'fully qualified list name', i.e. the posting address of the mailing
    list is required.  It must be a valid email address and the domain must be
    registered with Mailman.  List names are forced to lower case."""))
@click.option(
    '--language', metavar='CODE',
    help=_("""\
    Set the list's preferred language to CODE, which must be a registered two
    letter language code."""))
@click.option(
    '--owner', '-o', 'owners',
    multiple=True, metavar='OWNER',
    help=_("""\
    Specify a list owner email address.  If the address is not currently
    registered with Mailman, the address is registered and linked to a user.
    Mailman will send a confirmation message to the address, but it will also
    send a list creation notice to the address.  More than one owner can be
    specified."""))
@click.option(
    '--notify/-no-notify', '-n/-N',
    default=False,
    help=_("""\
    Notify the list owner by email that their mailing list has been
    created."""))
@click.option(
    '--quiet', '-q',
    is_flag=True, default=False,
    help=_('Print less output.'))
@click.option(
    '--domain/--no-domain', '-d/-D', 'create_domain',
    default=True,
    help=_("""\
    Register the mailing list's domain if not yet registered.  This is
    the default behavior, but these options are provided for backward
    compatibility.  With -D do not register the mailing list's domain."""))
@click.argument('fqdn_listname', metavar='LISTNAME')
@click.pass_context
def create(ctx, language, owners, notify, quiet, create_domain, fqdn_listname):
    language_code = (language if language is not None
                     else system_preferences.preferred_language.code)
    # Make sure that the selected language code is known.
    if language_code not in getUtility(ILanguageManager).codes:
        ctx.fail(_('Invalid language code: ${language_code}'))
    # Check to see if the domain exists or not.
    listname, at, domain = fqdn_listname.partition('@')
    domain_manager = getUtility(IDomainManager)
    if domain_manager.get(domain) is None and create_domain:
        domain_manager.add(domain)
    # Validate the owner email addresses.  The problem with doing this check in
    # create_list() is that you wouldn't be able to distinguish between an
    # InvalidEmailAddressError for the list name or the owners.  I suppose we
    # could subclass that exception though.
    if len(owners) > 0:
        validator = getUtility(IEmailValidator)
        invalid_owners = [owner for owner in owners
                          if not validator.is_valid(owner)]
        if invalid_owners:
            invalid = COMMASPACE.join(sorted(invalid_owners))  # noqa: F841
            ctx.fail(_('Illegal owner addresses: ${invalid}'))
    try:
        mlist = create_list(fqdn_listname, owners)
    except InvalidEmailAddressError:
        ctx.fail(_('Illegal list name: ${fqdn_listname}'))
    except ListAlreadyExistsError:
        ctx.fail(_('List already exists: ${fqdn_listname}'))
    except BadDomainSpecificationError as domain:              # noqa: F841
        ctx.fail(_('Undefined domain: ${domain}'))
    # Find the language associated with the code, then set the mailing list's
    # preferred language to that.
    language_manager = getUtility(ILanguageManager)
    with transaction():
        mlist.preferred_language = language_manager[language_code]
    # Do the notification.
    if not quiet:
        print(_('Created mailing list: ${mlist.fqdn_listname}'))
    if notify:
        template = getUtility(ITemplateLoader).get(
            'domain:admin:notice:new-list', mlist)
        text = wrap(expand(template, mlist, dict(
            # For backward compatibility.
            requestaddr=mlist.request_address,
            siteowner=mlist.no_reply_address,
            )))
        # Set the I18N language to the list's preferred language so the header
        # will match the template language.  Stashing and restoring the old
        # translation context is just (healthy? :) paranoia.
        with _.using(mlist.preferred_language.code):
            msg = UserNotification(
                owners, mlist.no_reply_address,
                _('Your new mailing list: ${fqdn_listname}'),
                text, mlist.preferred_language)
            msg.send(mlist)


@public
@implementer(ICLISubCommand)
class Create:
    name = 'create'
    command = create


@click.command(
    cls=I18nCommand,
    help=_('Remove a mailing list.'))
@click.option(
    '--quiet', '-q',
    is_flag=True, default=False,
    help=_('Suppress status messages'))
@click.argument('listspec')
def remove(quiet, listspec):
    mlist = getUtility(IListManager).get(listspec)
    if mlist is None:
        if not quiet:
            print(_('No such list matching spec: ${listspec}'))
            sys.exit(0)
    with transaction():
        remove_list(mlist)
        if not quiet:
            print(_('Removed list: ${listspec}'))


@public
@implementer(ICLISubCommand)
class Remove:
    name = 'remove'
    command = remove



"""roster_visibility

Revision ID: 15401063d4e3
Revises: b2e694dfde35
Create Date: 2019-01-20 20:45:50.773097

"""

# revision identifiers, used by Alembic.

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import exists_in_db, is_sqlite


revision = '15401063d4e3'
down_revision = 'b2e694dfde35'


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if not exists_in_db(
            op.get_bind(), 'mailinglist', 'member_roster_visibility'):
        op.add_column(                                       # pragma: nocover
            'mailinglist',
            sa.Column('member_roster_visibility', sa.Integer(), nullable=True))


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    if not is_sqlite(op.get_bind()):
        op.drop_column('mailinglist', 'member_roster_visibility')   # noqa: E501 # pragma: nocover



# Copyright (C) 2002-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The email command 'who'."""

import re

from email.utils import formataddr
from mailman.core.i18n import _
from mailman.interfaces.command import ContinueProcessing, IEmailCommand
from mailman.interfaces.member import DeliveryMode, DeliveryStatus
from mailman.model.roster import RosterVisibility
from public import public
from zope.interface import implementer


NEWLINE = '\n'


@public
@implementer(IEmailCommand)
class Who:
    """The email 'who' command."""

    name = 'who'
    argument_description = ('[delivery=<enabled|disabled>] '
                            '[mode=<digest|regular>]')
    description = _("""\
Produces a list of member names and email addresses.

The optional delivery= and mode= arguments can be used to limit the report
to those members with matching delivery status and/or delivery mode.  If
either delivery= or mode= is specified more than once, only the last occurrence
is used.
""")
    short_description = _('Get a list of the list members.')

    def process(self, mlist, msg, msgdata, arguments, results):
        """See `IEmailCommand`."""
        show_roster = mlist.member_roster_visibility == RosterVisibility.public
        show_roster |= (mlist.member_roster_visibility ==
                        RosterVisibility.members and
                        (bool(mlist.members.get_member(msg.sender)) or
                         bool(mlist.administrators.get_member(msg.sender))))
        show_roster |= (mlist.member_roster_visibility ==
                        RosterVisibility.moderators and
                        bool(mlist.administrators.get_member(msg.sender)))
        if not show_roster:
            print(_('You are not authorized to see the membership list.'),
                  file=results)
            return ContinueProcessing.no
        delivery = mode = None
        unrecognized = list()
        for arg in arguments:
            mo_deliv = re.match('^delivery=(enabled|disabled)$', arg)
            if mo_deliv is not None:
                delivery = mo_deliv.group(1)
            mo_mode = re.match('^mode=(digest|regular)$', arg)
            if mo_mode is not None:
                mode = mo_mode.group(1)
            if mo_deliv is None and mo_mode is None:
                unrecognized.append(arg)
        if len(unrecognized) > 0:
            print(_('Unrecognized or invalid argument(s):\n{}').format(
                  NEWLINE.join(unrecognized)), file=results)
            return ContinueProcessing.no
        members = list()
        for member in mlist.members.members:
            if delivery is not None:
                if (delivery == 'enabled' and
                        member.delivery_status != DeliveryStatus.enabled):
                    continue
                elif (delivery != 'enabled' and
                        member.delivery_status == DeliveryStatus.enabled):
                    continue
            if mode is not None:
                if (mode == 'regular' and
                        member.delivery_mode != DeliveryMode.regular):
                    continue
                elif (mode != 'regular' and
                        member.delivery_mode == DeliveryMode.regular):
                    continue
            members.append((member.display_name, member.address.email))
        members.sort(key=lambda x: x[1])
        for i in range(len(members)):
            members[i:i+1] = ['    ' + formataddr(members[i])]
        print(_('Members of the {} mailing list:\n{}').format(
              mlist.fqdn_listname, NEWLINE.join(members)), file=results)
        return ContinueProcessing.yes



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Database factory."""

import os
import types
import alembic.command

from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from flufl.lock import Lock
from mailman.config import config
from mailman.database.alembic import alembic_cfg
from mailman.database.model import Model
from mailman.interfaces.database import (
    DatabaseError,
    IDatabase,
    IDatabaseFactory,
)
from mailman.utilities.modules import call_name
from public import public
from sqlalchemy import MetaData
from zope.interface import implementer
from zope.interface.verify import verifyObject


LAST_STORM_SCHEMA_VERSION = '20130406000000'


@public
@implementer(IDatabaseFactory)
class DatabaseFactory:
    """Create a new database."""

    @staticmethod
    def create():
        """See `IDatabaseFactory`."""
        with Lock(os.path.join(config.LOCK_DIR, 'dbcreate.lck')):
            database_class = config.database['class']
            database = call_name(database_class)
            verifyObject(IDatabase, database)
            database.initialize()
            SchemaManager(database).setup_database()
            database.commit()
            return database


@public
class SchemaManager:
    """Manage schema migrations."""

    def __init__(self, database):
        self._database = database
        self._script = ScriptDirectory.from_config(alembic_cfg)

    def _get_storm_schema_version(self):
        metadata = MetaData()
        metadata.reflect(bind=self._database.engine)
        if 'version' not in metadata.tables:
            # There are no Storm artifacts left.
            return None
        Version = metadata.tables['version']
        last_version = self._database.store.query(Version.c.version).filter(
            Version.c.component == 'schema'
            ).order_by(Version.c.version.desc()).first()
        # Don't leave open transactions or they will block any schema change.
        self._database.commit()
        return last_version

    def setup_database(self):
        context = MigrationContext.configure(self._database.store.connection())
        current_rev = context.get_current_revision()
        head_rev = self._script.get_current_head()
        if current_rev == head_rev:
            # We're already at the latest revision so there's nothing to do.
            return head_rev
        if current_rev is None:
            # No Alembic information is available.
            storm_version = self._get_storm_schema_version()
            if storm_version is None:
                # Initial database creation.
                Model.metadata.create_all(self._database.engine)
                self._database.commit()
                alembic.command.stamp(alembic_cfg, 'head')
            else:
                # The database was previously managed by Storm.
                if storm_version.version < LAST_STORM_SCHEMA_VERSION:
                    raise DatabaseError(
                        'Upgrades skipping beta versions is not supported.')
                # Run migrations to remove the Storm-specific table and upgrade
                # to SQLAlchemy and Alembic.
                alembic.command.upgrade(alembic_cfg, 'head')
        elif current_rev != head_rev:
            alembic.command.upgrade(alembic_cfg, 'head')
        return head_rev


def _reset(self):
    """See `IDatabase`."""
    # Avoid a circular import at module level.
    from mailman.database.model import Model
    self.store.expunge_all()
    self.abort()
    self._pre_reset(self.store)
    Model._reset(self)
    self._post_reset(self.store)
    self.commit()


@public
@implementer(IDatabaseFactory)
class DatabaseTestingFactory:
    """Create a new database for testing."""

    @staticmethod
    def create():
        """See `IDatabaseFactory`."""
        database_class = config.database['class']
        database = call_name(database_class)
        verifyObject(IDatabase, database)
        database.initialize()
        # Remove existing tables (PostgreSQL will keep them across runs)
        metadata = MetaData()
        metadata.reflect(bind=database.engine)
        metadata.drop_all(bind=database.engine)
        database.commit()
        # Now create the current model without Alembic upgrades.
        Model.metadata.create_all(database.engine)
        database.commit()
        # Make _reset() a bound method of the database instance.
        database._reset = types.MethodType(_reset, database)
        return database



# Copyright (C) 2020-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Use utf8mb4 in MySQL

Revision ID: bc0c49c6dda2
Revises: ec5fe422e27c
Create Date: 2021-05-17 15:07:36.808010

"""
from alembic import op
from mailman.database.helpers import is_mysql
from mailman.database.types import SAUnicode4Byte, SAUnicodeLarge
from sqlalchemy.dialects import mysql


# revision identifiers, used by Alembic.
revision = 'bc0c49c6dda2'
down_revision = 'ec5fe422e27c'


def upgrade():  # pragma: nocover
    if not is_mysql(op.get_bind()):
        return
    with op.batch_alter_table('address', schema=None) as batch_op:
        batch_op.alter_column(
            'display_name',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)

    with op.batch_alter_table('domain', schema=None) as batch_op:
        batch_op.alter_column(
            'description',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)

    with op.batch_alter_table('headermatch', schema=None) as batch_op:
        batch_op.alter_column(
            'pattern',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)

    with op.batch_alter_table('mailinglist', schema=None) as batch_op:
        batch_op.alter_column(
            'autoresponse_owner_text',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'autoresponse_postings_text',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'autoresponse_request_text',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'bounce_matching_headers',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'dmarc_moderation_notice',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=510),
            type_=SAUnicodeLarge(),
            existing_nullable=True)
        batch_op.alter_column(
            'dmarc_wrapped_message_text',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=510),
            type_=SAUnicodeLarge(),
            existing_nullable=True)
        batch_op.alter_column(
            'description',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'member_moderation_notice',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'nonmember_rejection_notice',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'display_name',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'subject_prefix',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)

    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.alter_column(
            'display_name',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)
        batch_op.alter_column(
            'password',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=255),
            type_=SAUnicode4Byte(),
            existing_nullable=True)

    with op.batch_alter_table('workflowstate', schema=None) as batch_op:
        batch_op.alter_column(
            'data',
            existing_type=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                        length=510),
            type_=SAUnicodeLarge(),
            existing_nullable=True)


def downgrade():  # pragma: nocover
    if not is_mysql(op.get_bind()):
        return
    with op.batch_alter_table('workflowstate', schema=None) as batch_op:
        batch_op.alter_column(
            'data',
            existing_type=SAUnicodeLarge(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=510),
            existing_nullable=True)

    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.alter_column(
            'password',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'display_name',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)

    with op.batch_alter_table('mailinglist', schema=None) as batch_op:
        batch_op.alter_column(
            'subject_prefix',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'display_name',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'nonmember_rejection_notice',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'member_moderation_notice',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'description',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'dmarc_wrapped_message_text',
            existing_type=SAUnicodeLarge(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=510),
            existing_nullable=True)
        batch_op.alter_column(
            'dmarc_moderation_notice',
            existing_type=SAUnicodeLarge(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=510),
            existing_nullable=True)
        batch_op.alter_column(
            'bounce_matching_headers',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'autoresponse_request_text',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'autoresponse_postings_text',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)
        batch_op.alter_column(
            'autoresponse_owner_text',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)

    with op.batch_alter_table('headermatch', schema=None) as batch_op:
        batch_op.alter_column(
            'pattern',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)

    with op.batch_alter_table('domain', schema=None) as batch_op:
        batch_op.alter_column(
            'description',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)

    with op.batch_alter_table('address', schema=None) as batch_op:
        batch_op.alter_column(
            'display_name',
            existing_type=SAUnicode4Byte(),
            type_=mysql.VARCHAR(charset='utf8', collation='utf8_bin',
                                length=255),
            existing_nullable=True)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Subscription services."""

from mailman.app.membership import delete_member
from mailman.database.transaction import dbconnection
from mailman.interfaces.listmanager import IListManager, NoSuchListError
from mailman.interfaces.member import MemberRole
from mailman.interfaces.subscriptions import (
    ISubscriptionService,
    TooManyMembersError,
)
from mailman.interfaces.usermanager import IUserManager
from mailman.model.address import Address
from mailman.model.member import Member
from mailman.model.user import User
from mailman.utilities.queries import QuerySequence
from operator import attrgetter
from public import public
from sqlalchemy import or_, select, union_all
from sqlalchemy.orm import aliased
from zope.component import getUtility
from zope.interface import implementer


EMPTY = object()


@public
@implementer(ISubscriptionService)
class SubscriptionService:
    """Subscription services for the REST API."""

    __name__ = 'members'

    def get_members(self):
        """See `ISubscriptionService`."""
        # {list_id -> {role -> [members]}}
        by_list = {}
        user_manager = getUtility(IUserManager)
        for member in user_manager.members:
            by_role = by_list.setdefault(member.list_id, {})
            members = by_role.setdefault(member.role.name, [])
            members.append(member)
        # Flatten into single list sorted as per the interface.
        all_members = []
        address_of_member = attrgetter('address.email')
        for list_id in sorted(by_list):
            by_role = by_list[list_id]
            all_members.extend(
                sorted(by_role.get('owner', []), key=address_of_member))
            all_members.extend(
                sorted(by_role.get('moderator', []), key=address_of_member))
            all_members.extend(
                sorted(by_role.get('member', []), key=address_of_member))
        return all_members

    @dbconnection
    def get_member(self, store, member_id):
        """See `ISubscriptionService`."""
        members = store.query(Member).filter(Member._member_id == member_id)
        if members.count() == 0:
            return None
        else:
            assert members.count() == 1, 'Too many matching members'
            return members[0]

    @dbconnection
    def _find_members(self, store, subscriber, list_id, role,
                      delivery_mode, moderation_action, delivery_status):
        # If `subscriber` is a user id, then we'll search for all addresses
        # which are controlled by the user, otherwise we'll just search for
        # the given address.
        if (subscriber is None and
                list_id is None and
                role is None and
                delivery_mode is None and
                moderation_action is EMPTY and
                delivery_status is None):
            return []
        order = (Member.list_id, Address.email, Member.role)
        # Querying for the subscriber is the most complicated part, because
        # the parameter can either be an email address or a user id.  Start by
        # building two queries, one joined on the member's address, and one
        # joined on the member's user.  Add the resulting email address to the
        # selected values to be able to sort on it later on.
        q_address = select(Member, Address.email).join(Member._address)
        q_user = select(Member, Address.email).join(
            User, User.id == Member.user_id).join(User._preferred_address)
        if subscriber is not None:
            if isinstance(subscriber, str):
                # subscriber is an email address.
                subscriber = subscriber.lower()
                if '*' in subscriber:
                    subscriber = subscriber.replace('*', '%')
                    q_address = q_address.filter(or_(
                        Address.email.like(subscriber),
                        Address.display_name.ilike(subscriber)))
                    q_user = q_user.filter(or_(
                        Address.email.like(subscriber),
                        User.display_name.ilike(subscriber)))
                else:
                    q_address = q_address.filter(Address.email == subscriber)
                    q_user = q_user.filter(Address.email == subscriber)
            else:
                # subscriber is a user id.
                q_address = q_address.join(Address.user).filter(
                    User._user_id == subscriber)
                q_user = q_user.filter(User._user_id == subscriber)
        else:
            # We're not searching for a subscriber so only select preferred
            # addresses (see GL issue 227).
            q_user = q_user.filter(Address.id == User._preferred_address_id)
        # Add additional filters to both queries.
        if list_id is not None:
            q_address = q_address.filter(Member.list_id == list_id)
            q_user = q_user.filter(Member.list_id == list_id)
        if role is not None:
            q_address = q_address.filter(Member.role == role)
            q_user = q_user.filter(Member.role == role)
        if moderation_action is not EMPTY:
            q_address = q_address.filter(
                Member.moderation_action == moderation_action)
            q_user = q_user.filter(
                Member.moderation_action == moderation_action)
        # Do a UNION of the two queries, sort the result and generate Members.
        union = union_all(q_address, q_user).order_by(*order)
        stmt = select(aliased(Member, union.subquery()))
        query = QuerySequence(store, stmt)
        # These are python attributes and not actual table colukmsn and hence
        # their filtering can only be done in Python.
        if delivery_status or delivery_mode:
            return self._filter(query, delivery_mode, delivery_status)
        return query

    def _filter(self, query, delivery_mode, delivery_status):
        if delivery_mode is not None:
            query = filter(lambda m: m.delivery_mode == delivery_mode, query)
        if delivery_status is not None:
            query = filter(
                lambda m: m.delivery_status == delivery_status, query)
        return list(query)

    def find_members(self, subscriber=None, list_id=None, role=None,
                     delivery_mode=None, moderation_action=EMPTY,
                     delivery_status=None):
        """See `ISubscriptionService`."""
        return self._find_members(
            subscriber, list_id, role, delivery_mode, moderation_action,
            delivery_status)

    def find_member(self, subscriber=None, list_id=None, role=None):
        """See `ISubscriptionService`."""
        result = self._find_members(
            subscriber, list_id, role, None, EMPTY, None)
        if len(result) == 0:
            return None
        elif len(result) == 1:
            return result[0]
        raise TooManyMembersError(subscriber, list_id, role)

    def __iter__(self):
        yield from self.get_members()

    def leave(self, list_id, email):
        """See `ISubscriptionService`."""
        mlist = getUtility(IListManager).get_by_list_id(list_id)
        if mlist is None:
            raise NoSuchListError(list_id)
        # XXX for now, no notification or user acknowledgment.
        delete_member(mlist, email, False, False)

    @dbconnection
    def unsubscribe_members(self, store, list_id, emails):
        """See 'ISubscriptionService'."""
        success = set()
        fail = set()
        mlist = getUtility(IListManager).get_by_list_id(list_id)
        if mlist is None:
            raise NoSuchListError(list_id)
        # Start with a query on the matching list-id and role.
        q_member = store.query(Member).filter(
            Member.list_id == list_id,
            Member.role == MemberRole.member)
        # De-duplicate.
        for email in set(emails):
            unsubscribed = False
            # Join with a queries matching the email address and preferred
            # address of any subscribed user.
            q_address = q_member.join(Member._address).filter(
                Address.email == email)
            q_user = q_member.join(Member._user).join(
                User._preferred_address).filter(Address.email == email)
            members = q_address.union(q_user).all()
            for member in members:
                member.unsubscribe()
                unsubscribed = True
            (success if unsubscribed else fail).add(email)
        return success, fail



# Copyright (C) 2019-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Utilities to retry on failure."""

from functools import wraps


def retry(ExceptionToCheck, tries=1):
    """Retry calling the decorated function.

    Original from
    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries = tries
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck:
                    mtries -= 1
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST for users."""

from functools import lru_cache
from lazr.config import as_boolean
from mailman.config import config
from mailman.interfaces.address import (
    ExistingAddressError,
    InvalidEmailAddressError,
)
from mailman.interfaces.usermanager import IUserManager
from mailman.rest.addresses import PreferredAddress, UserAddresses
from mailman.rest.helpers import (
    bad_request,
    BadRequest,
    child,
    CollectionMixin,
    conflict,
    created,
    etag,
    forbidden,
    GetterSetter,
    no_content,
    not_found,
    NotFound,
    okay,
)
from mailman.rest.preferences import Preferences
from mailman.rest.validator import (
    list_of_strings_validator,
    PatchValidator,
    ReadOnlyPATCHRequestError,
    UnknownPATCHRequestError,
    Validator,
)
from passlib.utils import generate_password as generate
from public import public
from zope.component import getUtility


# Attributes of a user which can be changed via the REST API.
@public
class PasswordEncrypterGetterSetter(GetterSetter):
    def __init__(self):
        super().__init__(config.password_context.encrypt)

    def get(self, obj, attribute):
        assert attribute == 'cleartext_password'
        super().get(obj, 'password')

    def put(self, obj, attribute, value):
        assert attribute == 'cleartext_password'
        super().put(obj, 'password', value)


@public
class ListOfDomainOwners(GetterSetter):
    def get(self, domain, attribute):
        assert attribute == 'owner', (
            'Unexpected attribute: {}'.format(attribute))
        def sort_key(owner):                                      # noqa: E306
            return owner.addresses[0].email
        return sorted(domain.owners, key=sort_key)

    def put(self, domain, attribute, value):
        assert attribute == 'owner', (
            'Unexpected attribute: {}'.format(attribute))
        domain.add_owners(value)


ATTRIBUTES = dict(
    cleartext_password=PasswordEncrypterGetterSetter(),
    display_name=GetterSetter(str),
    is_server_owner=GetterSetter(as_boolean),
    )


CREATION_FIELDS = dict(
    display_name=str,
    email=str,
    is_server_owner=as_boolean,
    password=str,
    _optional=('display_name', 'password', 'is_server_owner'),
    )


def create_user(api, arguments, response):
    """Create a new user."""
    # We can't pass the 'password' argument to the user creation method, so
    # strip that out (if it exists), then create the user, adding the password
    # after the fact if successful.
    password = arguments.pop('password', None)
    is_server_owner = arguments.pop('is_server_owner', False)
    user_manager = getUtility(IUserManager)
    try:
        user = user_manager.create_user(**arguments)
    except ExistingAddressError as error:
        # The address already exists.  If the address already has a user
        # linked to it, raise an error, otherwise create a new user and link
        # it to this address.
        email = arguments.pop('email')
        user = user_manager.get_user(email)
        if user is None:
            address = user_manager.get_address(email)
            user = user_manager.create_user(**arguments)
            user.link(address)
        else:
            bad_request(
                response,
                'User already exists: {}'.format(error.address).encode())
            return None
    if password is None:
        # This will have to be reset since it cannot be retrieved.
        password = generate(int(config.passwords.password_length))
    user.password = config.password_context.encrypt(password)
    user.is_server_owner = is_server_owner
    user_id = api.from_uuid(user.user_id)
    location = api.path_to('users/{}'.format(user_id))
    created(response, location)
    return user


class _UserBase(CollectionMixin):
    """Shared base class for user representations."""

    def _resource_as_dict(self, user):
        """See `CollectionMixin`."""
        # The canonical URL for a user is their unique user id, although we
        # can always look up a user based on any registered and validated
        # email address associated with their account.  The user id is a UUID,
        # but we serialize its integer equivalent.
        user_id = self.api.from_uuid(user.user_id)
        resource = dict(
            created_on=user.created_on,
            is_server_owner=user.is_server_owner,
            self_link=self.api.path_to('users/{}'.format(user_id)),
            user_id=user_id,
            )
        # Add the password attribute, only if the user has a password.  Same
        # with the real name.  These could be None or the empty string.
        if user.password:
            resource['password'] = user.password
        if user.display_name:
            resource['display_name'] = user.display_name
        return resource

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return getUtility(IUserManager).users


@public
class AllUsers(_UserBase):
    """The users."""

    def on_get(self, request, response):
        """/users"""
        resource = self._make_collection(request)
        okay(response, etag(resource))

    def on_post(self, request, response):
        """Create a new user."""
        try:
            validator = Validator(**CREATION_FIELDS)
            arguments = validator(request)
            create_user(self.api, arguments, response)
        except (ValueError, InvalidEmailAddressError) as error:
            error_message = 'Invalid email address {}'.format(error)
            bad_request(response, error_message)
            return


class _FoundUsers(_UserBase):
    def __init__(self, users, api):
        super().__init__()
        self._users = users
        self.api = api

    def _get_collection(self, request):
        return self._users


class FindUsers(_UserBase):

    def on_get(self, request, response):
        validator = Validator(q=str,
                              # Allow pagination.
                              page=int,
                              count=int,
                              _optional=('page', 'count'))
        try:
            data = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return

        user_manager = getUtility(IUserManager)
        users = list(user_manager.find_users(data.get('q')))
        resource = _FoundUsers(users, self.api)
        try:
            collection = resource._make_collection(request)
        except ValueError as ex:
            bad_request(response, str(ex))
        else:
            okay(response, etag(collection))


@public
class AUser(_UserBase):
    """A user."""

    def __init__(self, user_identifier):
        """Get a user by various type of identifiers.

        :param user_identifier: The identifier used to retrieve the user.  The
            identifier may either be an email address controlled by the user
            or the UUID of the user.  The type of identifier is auto-detected
            by looking for an `@` symbol, in which case it's taken as an email
            address, otherwise it's assumed to be a UUID.  However, UUIDs in
            API 3.0 are integers, while in 3.1 are hex.
        :type user_identifier: string
        """
        self._user_identifier = user_identifier
        # Defer calculation of the user until the API object is set, since
        # that will determine how to interpret the user identifier.  For ease
        # of code migration, use an _user caching property (see below).

    @property
    @lru_cache(1)
    def _user(self):
        user_manager = getUtility(IUserManager)
        if '@' in self._user_identifier:
            return user_manager.get_user(self._user_identifier)
        else:
            # The identifier is the string representation of a UUID, either an
            # int in API 3.0 or a hex in API 3.1.
            try:
                user_id = self.api.to_uuid(self._user_identifier)
            except ValueError:
                return None
            else:
                return user_manager.get_user_by_id(user_id)

    def on_get(self, request, response):
        """Return a single user end-point."""
        if self._user is None:
            not_found(response)
        else:
            okay(response, self._resource_as_json(self._user))

    @child()
    def addresses(self, context, segments):
        """/users/<uid>/addresses"""
        if self._user is None:
            return NotFound(), []
        return UserAddresses(self._user)

    @child()
    def preferred_address(self, context, segements):
        """Return a user's preferred address."""
        if self._user is None:
            return NotFound(), []
        return PreferredAddress(self._user)

    def on_delete(self, request, response):
        """Delete the named user and all associated resources."""
        if self._user is None:
            not_found(response)
            return
        for member in self._user.memberships.members:
            member.unsubscribe()
        user_manager = getUtility(IUserManager)
        user_manager.delete_user(self._user)
        no_content(response)

    @child()
    def preferences(self, context, segments):
        """/users/<id>/preferences"""
        if len(segments) != 0:
            return BadRequest(), []
        if self._user is None:
            return NotFound(), []
        child = Preferences(
            self._user.preferences,
            'users/{}'.format(self.api.from_uuid(self._user.user_id)))
        return child, []

    def on_patch(self, request, response):
        """Patch the user's configuration (i.e. partial update)."""
        if self._user is None:
            not_found(response)
            return
        try:
            validator = PatchValidator(request, ATTRIBUTES)
        except UnknownPATCHRequestError as error:
            bad_request(
                response,
                'Unknown attribute: {0}'.format(error.attribute).encode())
        except ReadOnlyPATCHRequestError as error:
            bad_request(
                response,
                'Read-only attribute: {0}'.format(error.attribute).encode())
        else:
            validator.update(self._user, request)
            no_content(response)

    def on_put(self, request, response):
        """Put the user's configuration (i.e. full update)."""
        if self._user is None:
            not_found(response)
            return
        validator = Validator(**ATTRIBUTES)
        try:
            validator.update(self._user, request)
        except UnknownPATCHRequestError as error:
            bad_request(
                response,
                'Unknown attribute: {0}'.format(error.attribute).encode())
        except ReadOnlyPATCHRequestError as error:
            bad_request(
                response,
                'Read-only attribute: {0}'.format(error.attribute).encode())
        except ValueError as error:
            bad_request(response, str(error))
        else:
            no_content(response)

    @child()
    def login(self, context, segments):
        """Log the user in, sort of, by verifying a given password."""
        if self._user is None:
            return NotFound(), []
        return Login(self._user)


@public
class AddressUser(_UserBase):
    """The user linked to an address."""

    def __init__(self, address):
        self._address = address
        self._user = address.user

    def on_get(self, request, response):
        """Return a single user end-point."""
        if self._user is None:
            not_found(response)
        else:
            okay(response, self._resource_as_json(self._user))

    def on_delete(self, request, response):
        """Delete the named user, all her memberships, and addresses."""
        if self._user is None:
            not_found(response)
            return
        self._user.unlink(self._address)
        no_content(response)

    def on_post(self, request, response):
        """Link a user to the address, and create it if needed."""
        if self._user:
            conflict(response)
            return
        # When creating a linked user by POSTing, the user either must already
        # exist, or it can be automatically created, if the auto_create flag
        # is given and true (if missing, it defaults to true).  However, in
        # this case we do not accept 'email' as a POST field.
        fields = CREATION_FIELDS.copy()
        del fields['email']
        fields['user_id'] = self.api.to_uuid
        fields['auto_create'] = as_boolean
        fields['_optional'] = fields['_optional'] + (
            'user_id', 'auto_create', 'is_server_owner')
        try:
            validator = Validator(**fields)
            arguments = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        user_manager = getUtility(IUserManager)
        if 'user_id' in arguments:
            user_id = arguments['user_id']
            user = user_manager.get_user_by_id(user_id)
            if user is None:
                bad_request(response, 'No user with ID {}'.format(
                    self.api.from_uuid(user_id)).encode())
                return
            okay(response)
        else:
            auto_create = arguments.pop('auto_create', True)
            if auto_create:
                # This sets the 201 or 400 status.
                user = create_user(self.api, arguments, response)
                if user is None:
                    return
            else:
                forbidden(response)
                return
        user.link(self._address)

    def on_put(self, request, response):
        """Set or replace the addresses's user."""
        if self._user:
            self._user.unlink(self._address)
        # Process post data and check for an existing user.
        fields = CREATION_FIELDS.copy()
        fields['user_id'] = self.api.to_uuid
        fields['_optional'] = fields['_optional'] + (
            'user_id', 'email', 'is_server_owner')
        try:
            validator = Validator(**fields)
            arguments = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        user_manager = getUtility(IUserManager)
        if 'user_id' in arguments:
            user_id = arguments['user_id']
            user = user_manager.get_user_by_id(user_id)
            if user is None:
                not_found(response,
                          'No user with ID {}'.format(user_id).encode())
                return
            okay(response)
        else:
            user = create_user(self.api, arguments, response)
            if user is None:
                return
        user.link(self._address)


@public
class Login:
    """<api>/users/<uid>/login"""

    def __init__(self, user):
        assert user is not None
        self._user = user

    def on_post(self, request, response):
        # We do not want to encrypt the plaintext password given in the POST
        # data.  That would hash the password, but we need to have the
        # plaintext in order to pass into passlib.
        validator = Validator(cleartext_password=GetterSetter(str))
        try:
            values = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        is_valid, new_hash = config.password_context.verify(
            values['cleartext_password'], self._user.password)
        if is_valid:
            if new_hash is not None:
                self._user.password = new_hash
            no_content(response)
        else:
            forbidden(response)


@public
class OwnersForDomain(_UserBase):
    """Owners for a particular domain."""

    def __init__(self, domain):
        self._domain = domain

    def on_get(self, request, response):
        """/domains/<domain>/owners"""
        if self._domain is None:
            not_found(response)
            return
        resource = self._make_collection(request)
        okay(response, etag(resource))

    def on_post(self, request, response):
        """POST to /domains/<domain>/owners """
        if self._domain is None:
            not_found(response)
            return
        validator = Validator(
            owner=ListOfDomainOwners(list_of_strings_validator))
        try:
            validator.update(self._domain, request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        return no_content(response)

    def on_delete(self, request, response):
        """DELETE to /domains/<domain>/owners"""
        if self._domain is None:
            not_found(response)
        try:
            # No arguments.
            Validator()(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        owner_email = [
            owner.addresses[0].email
            for owner in self._domain.owners
            ]
        for email in owner_email:
            self._domain.remove_owner(email)
        return no_content(response)

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return list(self._domain.owners)


@public
class ServerOwners(_UserBase):
    """All server owners."""

    def on_get(self, request, response):
        """/owners"""
        resource = self._make_collection(request)
        okay(response, etag(resource))

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return list(getUtility(IUserManager).server_owners)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interfaces for managing languages."""

from public import public
from zope.interface import Attribute, Interface


@public
class ILanguage(Interface):
    """The representation of a language."""

    code = Attribute('The 2-character language code.')

    charset = Attribute('The character set or encoding for this language.')

    description = Attribute("The language's description.")


@public
class ILanguageManager(Interface):
    """A language manager.

    Current, disabling languages is not supported.
    """

    def add(code, charset, description):
        """Teach the language manager about a language.

        :param code: The short two-character language code for the
            language.  If the language manager already knows about this code,
            the old language binding is lost.
        :type code: string
        :param charset: The character set that the language uses,
            e.g. 'us-ascii', 'iso-8859-1', or 'utf-8'
        :type charset: string
        :param description: The English description of the language,
            e.g. 'English' or 'French'.
        :type description: string
        :return: The language object just added.
        :rtype: ILanguage
        """

    codes = Attribute('An iterator over all known codes.')

    languages = Attribute('An iterator of all the languages.')

    def __getitem__(code):
        """Return the language associated with the language code.

        :param code: The 2-letter language code.
        :type code: string
        :return: The language instance.
        :rtype: `ILanguage`
        :raises KeyError: if code is not associated with a known language.
        """

    def get(code, default=None):
        """Return the language associated with the language code.

        :param code: The 2-letter language code.
        :type code: string
        :param default: The value to return if the code is not known.
        :type default: anything
        :return: The language instance or `default`.
        :rtype: `ILanguage` or `default`
        """

    def __contains__(code):
        """True if the language code is known.

        :return: A flag indicating whether the language code is known or not.
        :rtype: bool
        """

    def clear():
        """Remove all language code mappings."""



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Look for moderator pre-approval."""

import re

from email.iterators import typed_subpart_iterator
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.rules import IRule
from public import public
from zope.interface import implementer


EMPTYSTRING = ''
HEADERS = [
    'approve',
    'approved',
    'x-approve',
    'x-approved',
    ]


@public
@implementer(IRule)
class Approved:
    """Look for moderator pre-approval."""

    name = 'approved'
    description = _('The message has a matching Approve or Approved header.')
    record = True

    def _get_password(self, msg, missing):
        for header in HEADERS:
            password = msg.get(header, missing)
            if password is not missing:
                return password
        return missing

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        if mlist.moderator_password is None:
            return False
        # See if the message has an Approved or Approve header with a valid
        # moderator password.  Also look at the first non-whitespace line in
        # the file to see if it looks like an Approved header.
        missing = object()
        password = self._get_password(msg, missing)
        if password is missing:
            # Find the first text/plain part in the message
            part = None
            stripped = False
            payload = None
            for part in typed_subpart_iterator(msg, 'text', 'plain'):
                payload = part.get_payload(decode=True)
                break
            if payload is not None:
                charset = part.get_content_charset('us-ascii')
                try:
                    # Do the decoding inside the try/except so that if the
                    # charset is unknown, we'll just drop back to ascii.
                    payload = payload.decode(charset, 'replace')
                except LookupError:
                    # Unknown or empty charset.
                    payload = payload.decode('us-ascii', 'replace')
                line = ''
                lines = payload.splitlines(True)
                for lineno, line in enumerate(lines):
                    if line.strip() != '':
                        break
                if ':' in line:
                    header, value = line.split(':', 1)
                    if header.lower() in HEADERS:
                        password = value.strip()
                        # Now strip the first line from the payload so the
                        # password doesn't leak.
                        del lines[lineno]
                        reset_payload(part, EMPTYSTRING.join(lines))
                        stripped = True
            if stripped:
                # Now try all the text parts in case it's
                # multipart/alternative with the approved line in HTML or
                # other text part.  We make a pattern from the Approved line
                # and delete it from all text/* parts in which we find it.  It
                # would be better to just iterate forward, but email
                # compatability for pre Python 2.2 returns a list, not a true
                # iterator.
                #
                # This will process all the multipart/alternative parts in the
                # message as well as all other text parts.  We shouldn't find
                # the pattern outside the multipart/alternative parts, but if
                # we do, it is probably best to delete it anyway as it does
                # contain the password.
                #
                # Make a pattern to delete.  We can't just delete a line
                # because line of HTML or other fancy text may include
                # additional message text.  This pattern works with HTML.  It
                # may not work with rtf or whatever else is possible.
                pattern = (header + r':(\s|&nbsp;)*' + re.escape(password))
                for part in typed_subpart_iterator(msg, 'text'):
                    payload = part.get_payload(decode=True)
                    if payload is not None:
                        charset = part.get_content_charset('us-ascii')
                        try:
                            # Do the decoding inside the try/except so that if
                            # the charset is unknown, we'll just drop back to
                            # ascii.
                            payload = payload.decode(charset, 'replace')
                        except LookupError:
                            # Unknown or empty charset.
                            payload = payload.decode('us-ascii', 'replace')
                        if re.search(pattern, payload):
                            reset_payload(part, re.sub(pattern, '', payload))
        else:
            for header in HEADERS:
                del msg[header]
        if password is missing:
            return False
        # Email confirm command wants to know there was an Approved:.
        msgdata['has_approved'] = True
        # The following is a workaround for
        # https://foss.heptapod.net/python-libs/passlib/-/issues/133
        # Ensure the hash for `verify` is not bytes.
        if isinstance(mlist.moderator_password, str):
            mpw = mlist.moderator_password
        else:
            mpw = mlist.moderator_password.decode('utf-8')
        is_valid, new_hash = config.password_context.verify(password, mpw)
        if is_valid and new_hash:
            # Hash algorithm migration.
            mlist.moderator_password = new_hash
        return is_valid


def reset_payload(part, payload):
    # Set decoded payload maintaining content-type, charset, format and delsp.
    charset = part.get_content_charset() or 'us-ascii'
    content_type = part.get_content_type()
    format = part.get_param('format')
    delsp = part.get_param('delsp')
    del part['content-transfer-encoding']
    del part['content-type']
    try:
        part.set_payload(payload, charset)
    except LookupError:
        part.set_payload(payload, 'us-ascii')
    part.set_type(content_type)
    if format:
        part.set_param('Format', format)
    if delsp:
        part.set_param('DelSp', delsp)



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""MIME content filtering."""

from enum import Enum
from public import public
from zope.interface import Attribute, Interface


@public
class FilterAction(Enum):
    # Discard a message that matches the content type filter.
    discard = 0
    # Bounce the message back to the original author.
    bounce = 1
    # Discard and forward the message on to the list owner.
    forward = 2
    # Discard, but preserve it.
    preserve = 3


@public
class FilterType(Enum):
    # Filter MIME type.
    filter_mime = 0
    # Pass MIME type.
    pass_mime = 1
    # Filter file extension.
    filter_extension = 2
    # Pass file extension.
    pass_extension = 3


@public
class IContentFilter(Interface):
    """A single content filter settings for a mailing list."""

    mailing_list = Attribute(
        """The mailing list for this content filter.""")

    filter_pattern = Attribute(
        """The filter/pass content pattern.""")

    filter_type = Attribute(
        """Type of filter.""")



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST domain tests."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.database.transaction import transaction
from mailman.interfaces.domain import IDomainManager
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.template import ITemplateManager
from mailman.testing.helpers import call_api
from mailman.testing.layers import RESTLayer
from urllib.error import HTTPError
from zope.component import getUtility


class TestDomains(unittest.TestCase):
    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('test@example.com')

    def test_create_domains(self):
        # Create a domain with owners.
        data = dict(
            mail_host='example.org',
            description='Example domain',
            owner=['someone@example.com', 'secondowner@example.com'],
            )
        content, response = call_api(
            'http://localhost:9001/3.0/domains', data, method="POST")
        self.assertEqual(response.status_code, 201)

    def test_create_domain_bad_values(self):
        # Failed validation of values don't return 500 error but 400.
        data = dict(
            mail_host='example.org',
            random=10,
            )
        with self.assertRaises(HTTPError) as cm:
            content, response = call_api(
                'http://localhost:9001/3.0/domains', data, method="POST")
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason, 'Unexpected parameters: random')

    def test_patch_domain_description(self):
        # Patch the example.com description.
        data = {'description': 'Patched example domain'}
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com',
            data,
            method='PATCH')
        self.assertEqual(response.status_code, 204)
        # Check the result.
        domain = getUtility(IDomainManager).get('example.com')
        self.assertEqual(domain.description, 'Patched example domain')

    def test_patch_domain_multiple(self):
        # Patch the example.com alias_domain and description.
        data = dict(
            alias_domain='zzz.example.net',
            description='Some other description',
            )
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com',
            data,
            method='PATCH')
        self.assertEqual(response.status_code, 204)
        # Check the result.
        domain = getUtility(IDomainManager).get('example.com')
        self.assertEqual(domain.alias_domain, 'zzz.example.net')
        self.assertEqual(domain.description, 'Some other description')

    def test_patch_domain_owner(self):
        # Patch the example.com owner.
        data = {'owner': 'anne@example.com'}
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com',
            data,
            method='PATCH')
        self.assertEqual(response.status_code, 204)
        # Check the result.
        domain = getUtility(IDomainManager).get('example.com')
        self.assertEqual(
            [list(owner.addresses)[0].email for owner in domain.owners],
            ['anne@example.com'])

    def test_patch_domain_two_owners(self):
        # Patch the example.com owner.
        data = {'owner': ['anne@example.com', 'other@example.net']}
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com',
            data,
            method='PATCH')
        self.assertEqual(response.status_code, 204)
        # Check the result.
        domain = getUtility(IDomainManager).get('example.com')
        owners = [list(owner.addresses)[0].email for owner in domain.owners]
        owners.sort()
        self.assertEqual(owners, ['anne@example.com', 'other@example.net'])

    def test_patch_domain_readonly(self):
        # Attempt to patch mail_host.
        data = {'mail_host': 'example.net'}
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/domains/example.com',
                data,
                method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_domain_create_with_single_owner(self):
        # Creating domain with single owner should not raise InvalidEmailError.
        content, response = call_api(
            'http://localhost:9001/3.0/domains', dict(
                mail_host='example.net',
                owner='anne@example.com',
                ),
            method='POST')
        self.assertEqual(response.status_code, 201)
        # The domain has the expected owner.
        domain = getUtility(IDomainManager).get('example.net')
        self.assertEqual(
            [list(owner.addresses)[0].email for owner in domain.owners],
            ['anne@example.com'])

    def test_domain_create_with_alias(self):
        # Create a domain with alias_domain.
        content, response = call_api(
            'http://localhost:9001/3.0/domains', dict(
                mail_host='example.net',
                owner='anne@example.com',
                alias_domain='x.example.net',
                ),
            method='POST')
        self.assertEqual(response.status_code, 201)
        # The domain has the expected alias.
        domain = getUtility(IDomainManager).get('example.net')
        self.assertEqual(domain.alias_domain, 'x.example.net')

    def test_bogus_endpoint_extension(self):
        # /domains/<domain>/lists/<anything> is not a valid endpoint.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com'
                     '/lists/wrong')
        self.assertEqual(cm.exception.code, 400)

    def test_bogus_endpoint(self):
        # /domains/<domain>/<!lists> does not exist.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com/wrong')
        self.assertEqual(cm.exception.code, 404)

    def test_lists_are_deleted_when_domain_is_deleted(self):
        # /domains/<domain> DELETE removes all associated mailing lists.
        with transaction():
            create_list('ant@example.com')
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com', method='DELETE')
        self.assertEqual(response.status_code, 204)
        self.assertIsNone(getUtility(IListManager).get('ant@example.com'))

    def test_missing_domain(self):
        # You get a 404 if you try to access a nonexisting domain.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/does-not-exist.com')
        self.assertEqual(cm.exception.code, 404)

    def test_missing_domain_lists(self):
        # You get a 404 if you try to access the mailing lists of a
        # nonexisting domain.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/domains/does-not-exist.com/lists')
        self.assertEqual(cm.exception.code, 404)

    def test_create_existing_domain(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains', dict(
                mail_host='example.com',
                ))
        self.assertEqual(cm.exception.code, 400)

    def test_double_delete(self):
        # You cannot delete a domain twice.
        content, response = call_api(
            'http://localhost:9001/3.0/domains/example.com',
            method='DELETE')
        self.assertEqual(response.status_code, 204)
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)


class TestDomainOwners(unittest.TestCase):
    layer = RESTLayer

    def test_get_missing_domain_owners(self):
        # Try to get the owners of a missing domain.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.net/owners')
        self.assertEqual(cm.exception.code, 404)

    def test_bad_domain_owners_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/domains/example.com/owners/bogus')
        self.assertEqual(cm.exception.code, 404)

    def test_post_to_missing_domain_owners(self):
        # Try to add owners to a missing domain.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.net/owners', (
                ('owner', 'dave@example.com'), ('owner', 'elle@example.com'),
                ))
        self.assertEqual(cm.exception.code, 404)

    def test_delete_missing_domain_owners(self):
        # Try to delete the owners of a missing domain.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.net/owners',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)

    def test_bad_post(self):
        # Send POST data with an invalid attribute.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com/owners', (
                ('guy', 'dave@example.com'), ('gal', 'elle@example.com'),
                ))
        self.assertEqual(cm.exception.code, 400)

    def test_bad_delete(self):
        # Send DELETE with any data.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com/owners', {
                'owner': 'dave@example.com',
                }, method='DELETE')
        self.assertEqual(cm.exception.code, 400)


class TestDomainTemplates(unittest.TestCase):
    """Test /domains/<mail-host>/uris"""

    layer = RESTLayer

    def test_no_templates_for_api_30(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/example.com/uris')
        self.assertEqual(cm.exception.code, 404)

    def test_no_templates_for_missing_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/domains/example.org/uris')
        self.assertEqual(cm.exception.code, 404)

    def test_path_too_long(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/domains/example.com/uris'
                     '/foo/bar')
        self.assertEqual(cm.exception.code, 400)

    def test_get_unknown_uri(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/domains/example.com/uris'
                     '/not:a:template')
        self.assertEqual(cm.exception.code, 404)

    def test_get_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', 'example.com',
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', 'example.com',
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(resource['start'], 0)
        self.assertEqual(resource['total_size'], 2)
        self.assertEqual(
            resource['self_link'],
            'http://localhost:9001/3.1/domains/example.com/uris')
        self.assertEqual(resource['entries'], [
            {'http_etag': '"e877ff896db0f2e280660ac16b9401f7925a53b9"',
             'name': 'list:user:notice:goodbye',
             'password': 'the password',
             'self_link': ('http://localhost:9001/3.1/domains/example.com'
                           '/uris/list:user:notice:goodbye'),
             'uri': 'http://example.com/goodbye',
             'username': 'a user',
             },
            {'http_etag': '"8dac25601c3419e98e2c05df1d962a2252b67ce6"',
             'name': 'list:user:notice:welcome',
             'self_link': ('http://localhost:9001/3.1/domains/example.com'
                           '/uris/list:user:notice:welcome'),
             'uri': 'http://example.com/welcome',
             }])

    def test_patch_uris(self):
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')
        template = manager.raw('list:user:notice:goodbye', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')

    def test_patch_uris_with_credentials(self):
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_patch_uris_with_partial_credentials(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/domains/example.com/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'username': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_put_all_uris(self):
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris', {
                'domain:admin:notice:new-list': '',
                'list:admin:action:post': '',
                'list:admin:action:subscribe': '',
                'list:admin:action:unsubscribe': '',
                'list:admin:notice:subscribe': '',
                'list:admin:notice:unrecognized': '',
                'list:admin:notice:pending': '',
                'list:admin:notice:increment': '',
                'list:admin:notice:disable': '',
                'list:admin:notice:unsubscribe': '',
                'list:admin:notice:removal': '',
                'list:member:digest:footer': '',
                'list:member:digest:header': '',
                'list:member:digest:masthead': '',
                'list:member:regular:footer': 'http://example.org/footer',
                'list:member:regular:header': 'http://example.org/header',
                'list:user:action:invite': '',
                'list:user:action:subscribe': '',
                'list:user:action:unsubscribe': '',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'list:user:notice:hold': '',
                'list:user:notice:no-more-today': '',
                'list:user:notice:post': '',
                'list:user:notice:probe': '',
                'list:user:notice:refuse': '',
                'list:user:notice:rejected': '',
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:warning': '',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PUT')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:member:digest:footer', 'example.com')
        self.assertIsNone(template)
        template = manager.raw('list:member:digest:header', 'example.com')
        self.assertIsNone(template)
        template = manager.raw('list:member:regular:footer', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/footer')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:member:regular:header', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/header')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_delete_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', 'example.com',
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', 'example.com',
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris',
            method='DELETE')
        self.assertEqual(response.status_code, 204)
        self.assertIsNone(
            manager.raw('list:user:notice:welcome', 'example.com'))
        self.assertIsNone(
            manager.raw('list:user:notice:goodbye', 'example.com'))

    def test_get_a_url(self):
        with transaction():
            getUtility(ITemplateManager).set(
                'list:user:notice:welcome', 'example.com',
                'http://example.com/welcome')
        resource, response = call_api(
            'http://localhost:9001/3.1/domains/example.com/uris'
            '/list:user:notice:welcome')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(resource, {
            'http_etag': '"8884a0b3d675b4cb9899a7825daac9db88b70bed"',
            'self_link': ('http://localhost:9001/3.1/domains/example.com'
                          '/uris/list:user:notice:welcome'),
            'uri': 'http://example.com/welcome',
            })

    def test_get_a_bad_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/domains/example.com/uris'
                '/list:user:notice:notemplate')
        self.assertEqual(cm.exception.code, 404)

    def test_get_unset_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/domains/example.com/uris'
                '/list:user:notice:welcome')
        self.assertEqual(cm.exception.code, 404)

    def test_patch_url_with_too_many_parameters(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/domains/example.com/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'secret': 'some password',
                    'person': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)
